#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain.embeddings.base import Embeddings
from langchain_openai import OpenAIEmbeddings

from langchain_pangu.pangukitsappdev.api.embeddings.base import AbstractEmbeddingApi
from langchain_pangu.pangukitsappdev.api.embeddings.embedding_config import EmbeddingConfig


class OpenAIEmbeddingApi(AbstractEmbeddingApi):
    def create_embeddings(self, embedding_config: EmbeddingConfig) -> Embeddings:
        config_params = {}

        if embedding_config.openai_config.openai_base_url:
            config_params["openai_api_base"] = embedding_config.openai_config.openai_base_url

        if embedding_config.openai_config.openai_key:
            config_params["openai_api_key"] = embedding_config.openai_config.openai_key

        # 配置代理
        if embedding_config.openai_config.http_config.proxy_enabled:
            config_params["openai_proxy"] = embedding_config.openai_config.http_config.get_proxy_url()

        return OpenAIEmbeddings(**config_params)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain_pangu.pangukitsappdev.retriever.tool_metadata import ToolMetadata


class RetrievedTool(ToolMetadata):
    """
    工具的元数据信息
    Attributes:
        score: 评分，如从ES库检索后的score字段
    """
    score: float



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain.prompts import PromptTemplate

template = """请仔细阅读以下文章回答问题，不要遗漏任何信息，尽可能的覆盖全面，给出全面详细的回答。
如果问题与文章不相关，则无需参考文章内容，直接回答“抱歉我没有找到相关的参考资料。
文章:
=========
{% for item in documents %}{{ item.pageContent }}{% if not loop.last %},{% endif %}{% endfor %}
=========

问题：{{question}}
回答：
"""

input_variables = ["summaries", "question"]
PROMPT = PromptTemplate.from_template(template, template_format="jinja2")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import hashlib
import uuid
from typing import List, Tuple, Dict

import numpy as np
from elasticsearch.helpers import bulk
from gptcache import Cache, Config
from gptcache.adapter.api import init_similar_cache
from gptcache.embedding.langchain import LangChain
from gptcache.manager import get_data_manager
from gptcache.manager.scalar_data.redis_storage import RedisCacheStorage
from gptcache.manager.vector_data.base import VectorBase, VectorData
from langchain.cache import GPTCache

from langchain_pangu.pangukitsappdev.api.memory.cache.base import CacheApiAdapter
from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import CacheStoreConfig
from langchain_pangu.pangukitsappdev.api.memory.vector.base import Document
from langchain_pangu.pangukitsappdev.api.memory.vector.factory import Vectors
from langchain_pangu.pangukitsappdev.api.memory.vector.vector_config import VectorStoreConfig
from langchain_pangu.pangukitsappdev.vectorstores.bulk_data import BulkData
from langchain_pangu.pangukitsappdev.vectorstores.css_store import CSSVectorSearch

GPT_CACHE_CSS_INDEX_PREFIX = "gptcache_prompt_index"


def text_mapping(dim: int, metric: str = "inner_product") -> Dict:
    return {
        "mappings": {
            "properties": {
                "vector": {
                    "type": "vector",
                    "dimension": dim,
                    "indexing": 'true',
                    "algorithm": "GRAPH",
                    "metric": metric
                },
            }
        }, "settings": {"index": {"vector": "true"}}}


def get_doc_with_score_func(hit: dict) -> Tuple[Document, float]:
    return (
        Document(
            page_content=hit["_id"],
            metadata={},
            id=hit["_id"],
            score=hit["_score"]
        ),
        hit["_score"],
    )


def add_text_request(index_name: str,
                     vector: List[float],
                     content: str,
                     doc_metadata: dict) -> Tuple[str, dict]:
    """这里直接使用content作为es的id"""
    es_id = content
    req = {
        "_op_type": "index",
        "_index": index_name,
        "vector": vector,
        "_id": es_id,
    }

    return es_id, req


class CSSVectorStore(VectorBase):
    """
    适配GPTCache的VectorBase的类，用来集成进gptcache作为向量存储使用

    Attributes:
        _top_k: 默认的4，从CSS中检索出的文档数量
        css_vector_search: 内部分装的CSS向量存储
        vector_config: 用于构造实例的配置内容
    """

    def __init__(self, vector_config: VectorStoreConfig):
        """构造器
        Args:
            vector_config: 向量存储的配置
        """
        config = {
            "elasticsearch_url": vector_config.server_info.get_urls(),
            "index_name": vector_config.index_name,
            "embedding": vector_config.embedding,
            "verify_certs": vector_config.verify_certs,
            "text_key": vector_config.text_key,
            "vector_fields": vector_config.vector_fields,
            "source_fields": vector_config.source_fields,
            "mapping_func": text_mapping,
            "get_doc_with_score_func": get_doc_with_score_func,
        }

        self._top_k = 4
        self.css_vector_search = CSSVectorSearch(**config)
        self.vector_config = vector_config

    def mul_add(self, datas: List[VectorData]):
        bulk_data_list = [BulkData(id=uuid.uuid4(), datas={"texts": str(d.id)}) for d in datas]
        self.css_vector_search.add_docs_with_external_embedding(bulk_data_list=bulk_data_list,
                                                                embedded=[d.data.tolist() for d in datas])

    def search(self, data: np.ndarray, top_k: int):
        """
        相似检索
        Args:
            data: 向量数据
            top_k: 检索top_k个数据

        Returns:
            （score: float, id: int）

        """
        top_k = self._top_k if top_k == -1 else top_k
        docs_and_scores = self.css_vector_search.similarity_search_with_external_embedding(data.tolist(), top_k)
        return [(ds[1], int(ds[0].page_content)) for ds in docs_and_scores]

    def rebuild(self, ids=None) -> bool:
        return True

    def delete(self, ids) -> bool:
        # 构造批量删除请求
        bulk_request = []
        for doc_id in ids:
            delete_request = {
                "_op_type": "delete",  # 操作类型为删除
                "_index": self.vector_config.index_name,  # 要删除的文档所在的索引名称
                "_id": doc_id  # 要删除的文档ID
            }
            bulk_request.append(delete_request)

        # 执行批量删除请求
        bulk(self.css_vector_search.client, bulk_request)

    def flush(self):
        if not self.css_vector_search.client.indices.exists(self.vector_config.index_name):
            return
        self.css_vector_search.client.indices.flush(index=self.vector_config.index_name)


class SemanticGptCacheApi(CacheApiAdapter):
    """GptCache语义缓存
    基于GptCache组件做的适配和扩展。集成了CSS作为vector_store

    """

    def __init__(self, cache_config: CacheStoreConfig):
        def init_gptcache(cache_obj: Cache, llm: str):
            cache = RedisCacheStorage(url=cache_config.server_info.get_urls()[0])
            vector_config = VectorStoreConfig(
                index_name=f"{GPT_CACHE_CSS_INDEX_PREFIX}_{hashlib.sha256(llm.encode()).hexdigest()}",
                vector_fields=["vector"]
            )
            vector = Vectors.of(cache_config.vector_store_name, vector_config)
            data_manager = get_data_manager(cache_base=cache, vector_base=vector)
            init_similar_cache(cache_obj=cache_obj,
                               embedding=LangChain(cache_config.embedding),
                               data_manager=data_manager,
                               config=Config(similarity_threshold=cache_config.score_threshold))

        gpt_cache = GPTCache(init_gptcache)
        super().__init__(gpt_cache, cache_config.session_tag)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Type

from langchain_pangu.pangukitsappdev.api.memory.vector.base import VectorApi
from langchain_pangu.pangukitsappdev.api.memory.vector.vector_config import VectorStoreConfig
from langchain_pangu.pangukitsappdev.vectorstores.adapter import CSSVectorApi


class Vectors:
    vectors_map: Dict[str, Type[VectorApi]] = {}

    @classmethod
    def register(cls, vector_type: Type[VectorApi], vector_name: str):
        """注册一种vector的类型
        Args:
            vector_type: vector的类型，要求是BaseLLM的子类
            vector_name: vector的名字，唯一代表这个vector的名字
        """
        cls.vectors_map[vector_name] = vector_type

    @classmethod
    def of(cls, vector_name: str = "", vector_config: VectorStoreConfig = None) -> VectorApi:
        """根据名字创建一个VectorApi的实现类
        vector_name和vector_config.store_name不可同时为空
        Args:
            vector_name: （Optional）vector的名字，唯一标识一个VectorApi的实现类
            vector_config: （Optional）VectorApi的相关配置，如果不传递则从默认配置文件中或者环境变量中获取

        Returns:
            一个VectorApi的实现类

        """
        local_vector_name = vector_name if vector_name else vector_config.store_name

        if not local_vector_name:
            raise ValueError("The parameter vector_name or vector_config.store_name should have value")

        vector_type = cls.vectors_map.get(local_vector_name)
        if not vector_type:
            raise ValueError(
                f"Unregistered vector name: {local_vector_name}, \
                please call register(vector_type, vector_name) before use.")

        local_vector_config = vector_config if vector_config else cls._load_config()

        return vector_type(local_vector_config)

    @classmethod
    def _load_config(cls):
        return VectorStoreConfig()


Vectors.register(CSSVectorApi, "css")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import List

from langchain.schema import BaseRetriever, Document

from langchain_pangu.pangukitsappdev.utilities.web_search import WebSearch


class WebRetriever(BaseRetriever):
    web_search: WebSearch

    def _get_relevant_documents(self, query: str, top_k=5) -> List[Document]:
        return self.web_search.query(query, top_k)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import ABC, abstractmethod
from typing import List, Optional, Callable

from langchain_pangu.pangukitsappdev.api.llms.base import ConversationMessage
from langchain_pangu.pangukitsappdev.api.memory.vector.base import VectorApi, Document
from langchain_pangu.pangukitsappdev.api.memory.vector.vector_config import VectorStoreConfig
from langchain_pangu.pangukitsappdev.api.tool.base import AbstractTool
from langchain_pangu.pangukitsappdev.retriever.retrieved_tool import RetrievedTool
from langchain_pangu.pangukitsappdev.retriever.tool_metadata import ToolMetadata
from langchain_pangu.pangukitsappdev.tool.tool_provider import ToolProvider
from langchain_pangu.pangukitsappdev.vectorstores.bulk_data import BulkData


class ToolRetriever(ABC):
    """工具检索器接口类
    """

    @abstractmethod
    def add_tools(self, tools: List[AbstractTool]):
        """
        工具入库
        :param tools: tools工具
        """

    @abstractmethod
    def add_tools_from_metadata(self, tool_metadata_list: List[ToolMetadata]):
        """
        工具入库
        :param tool_metadata_list: 需要入库的工具
        """

    @abstractmethod
    def search(self, query: str,
               top_k: Optional[int] = None,
               score_threshold: Optional[float] = None) -> List[AbstractTool]:
        """
        工具检索
        :param query: 查询语句
        :param top_k: top k条
        :param score_threshold: 评分阈值
        :return: 相似工具列表
        """

    @abstractmethod
    def remove(self, tool_ids: List[str]):
        """
        删除工具
        :param tool_ids: tools工具
        """

    @abstractmethod
    def set_query_preprocessor(self, preprocessor: Callable[[List[ConversationMessage]], str]):
        """
        设置query请求的预处理器
        :param preprocessor: 当ToolRetriever被设置在Agent中时，Agent在Retrieve工具前，
        先会调用query预处理器对多轮对话进行改写，如果不设置，默认使用多轮中的最后一轮对话
        """

    @abstractmethod
    def get_query_preprocessor(self) -> Callable[[List[ConversationMessage]], str]:
        """
        获取query请求的预处理器
        :return: query预处理器
        """

    @abstractmethod
    def dsl_search(self, query: str, dsl: str) -> List[AbstractTool]:
        """
        工具检索
        :param query : 查询语句
        :param dsl : dsl
        :return: 相似工具列表
        """


class AbstractToolRetriever(ToolRetriever, ABC):
    RETRIEVER_SIZE_DEFAULT = 5
    RETRIEVER_SCORE_DEFAULT = 0.2

    def __init__(self, tool_provider: ToolProvider, vector_store_config: VectorStoreConfig):
        """储存工具类相关信息
        初始化
        :param tool_provider: 工具存储
        :param: vector_store_config: 语义向量存储配置
        """
        self.tool_provider: ToolProvider = tool_provider
        self.vector_store_config: VectorStoreConfig = vector_store_config
        self.vector: Optional[VectorApi] = None

        def default_processor(messages: List[ConversationMessage]) -> str:
            query = ""
            for message in reversed(messages):
                if message.role.text == "user":
                    query = message.content
                    break
            return query

        self.preprocessor: Callable[[List[ConversationMessage]], str] = default_processor

    def set_query_preprocessor(self, preprocessor: Callable[[List[ConversationMessage]], str]):
        self.preprocessor = preprocessor

    def get_query_preprocessor(self) -> Callable[[List[ConversationMessage]], str]:
        return self.preprocessor

    def add_tools_from_metadata(self, tool_metadata_list: List[ToolMetadata]):
        bulk_data_list = [BulkData(id=tool_metadata.tool_id,
                                   data=tool_metadata.tool_metadata) for tool_metadata in tool_metadata_list]
        self.vector.add_docs(bulk_data_list)

    def search(self, query: str,
               top_k: int = RETRIEVER_SIZE_DEFAULT,
               score_threshold: float = RETRIEVER_SCORE_DEFAULT) -> List[AbstractTool]:
        # 检索相关工具，获取id
        docs = self.vector.similarity_search(query=query,
                                             top_k=top_k,
                                             score_threshold=score_threshold)
        return self.instantiation_tool(docs, query)

    def remove(self, tool_ids: List[str]):
        self.vector.remove(tool_ids)

    def instantiation_tool(self, documents: List[Document], query: str) -> List[AbstractTool]:
        retrieved_tools = [RetrievedTool(score=doc.score,
                                         tool_id=doc.id,
                                         tool_metadata=doc.metadata) for doc in documents]
        return self.tool_provider.provide(retrieved_tools, query)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Any, List

from langchain.memory.chat_memory import BaseChatMemory

from langchain_pangu.pangukitsappdev.api.skill.base import SimpleSkill


class ConversationSummaryBufferMemory(BaseChatMemory):
    """Buffer with summarizer for storing conversation memory."""
    buffer: str = ""
    summary_skill: SimpleSkill
    human_prefix: str = "Human"
    ai_prefix: str = "AI"
    memory_key: str = "history"

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Return history buffer."""
        return {self.memory_key: self.buffer}

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save context from this conversation to buffer."""
        super().save_context(inputs, outputs)
        summary = {}
        summary.update(
            {"context": " ".join(["{}:{}".format(msg.type, msg.content) for msg in self.chat_memory.messages[-2:]])}
        )
        summary.update({"summary": self.buffer})
        self.buffer = self.summary_skill.execute(summary)

    def clear(self) -> None:
        """Clear memory contents."""
        super().clear()
        self.buffer = ""



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.
from pydantic.v1.utils import deepcopy

from langchain_pangu.pangukitsappdev.agent.agent_session import AgentSession
from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.api.skill.base import Skill, SimpleSkill
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates


class AgentSessionSkill(Skill):
    """
    对Agent的Session进行总结，得出问题的最终答案
    """

    def __init__(self, llm: LLMApi):
        self.llm = llm

    def summary(self, session: AgentSession) -> str:
        history_action = session.current_message.actions
        if not history_action:
            return ""
        # 如果只包含一轮，则直接返回
        if len(history_action) == 1:
            return session.current_message.content
        # 如果包含多轮则进行总结
        summary_skill = SimpleSkill(PromptTemplates.get("skill_agent_session_summary"), self.llm)
        messages = deepcopy(session.messages)
        if messages[-1].role.text == "assistant":
            messages.pop()
        return summary_skill.execute({"messages": session.messages, "actions": history_action})

    def execute(self, session: AgentSession) -> str:
        return self.summary(session)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import List

from langchain.callbacks.manager import CallbackManagerForRetrieverRun
from langchain.schema import Document
from langchain.schema.retriever import BaseRetriever

from langchain_pangu.pangukitsappdev.utilities.knowledge_graph import KnowledgeGraph


class KGRetriever(BaseRetriever):
    kg: KnowledgeGraph

    def _get_relevant_documents(self, query: str, *, run_manager: CallbackManagerForRetrieverRun) -> List[Document]:
        return self.kg.query(query)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from pydantic.v1 import Field

from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings


class PromptConfig(SdkBaseSettings):
    """Prompt参数
    Attributes:
        custom_prompt_path: 用户自定义prompt路径，优先生效
        default_prompt_path: 预制模板路径
    """
    custom_prompt_path: Optional[str] = Field(env="sdk.prompt.path")
    default_prompt_path: str = Field(default="prompts/default/")


class PromptTemplatesFileConfig(SdkBaseSettings):
    """模板文件存放结构
    Attributes:
        documents_stuff: documents/stuff.pt
        documents_refine_qa: documents/refine_qa.pt
        documents_refine_combine: documents/refine_combine.pt
        documents_mapreduce_map: documents/mapreduce_map.pt
        documents_mapreduce_reduce: documents/mapreduce_reduce.pt
        documents_summarize_stuff: documents/summarize_stuff.pt
        documents_summarize_refine_qa: documents/summarize_refine_qa.pt
        documents_summarize_refine_combine: documents/summarize_refine_combine.pt
        documents_summarize_mapreduce_map: documents/summarize_mapreduce_map.pt
        documents_summarize_mapreduce_reduce: documents/summarize_mapreduce_reduce.pt
        memory_summary: memory/summary.pt
        system_out_put_parser: system/out_put_parser.pt
        agent_react: agent/react.pt
        agent_react_pangu: agent/react_pangu.pt
        agent_tool_desc: agent/tool_desc.pt
        agent_tool_desc_pangu: agent/tool_desc_pangu.pt
        agent_tool_json_schema: agent/tool_json_schema.pt
        agent_react_pangu_2: agent/react_pangu_2.pt
        agent_tool_desc_pangu_2: agent/tool_desc_pangu_2.pt
        agent_react_pangu_unify: agent/react_pangu_unify.pt
        conversation_default: conversation/default.pt
        skill_conversation_rewrite: skill/conversation_rewrite.pt
        skill_agent_session_summary: skill/agent_session_summary.pt
        question_long_answer: question/long_answer.jinja2
        question_short_answer: question/short_answer.jinja2
        question_qa_with_sources_stuff: question/qa_with_sources_stuff.jinja2
        question_question_only: question/question_only.jinja2
    """
    documents_stuff: str = Field(default="documents/stuff.pt")
    documents_refine_qa: str = Field(default="documents/refine_qa.pt")
    documents_refine_combine: str = Field(default="documents/refine_combine.pt")
    documents_mapreduce_map: str = Field(default="documents/mapreduce_map.pt")
    documents_mapreduce_reduce: str = Field(default="documents/mapreduce_reduce.pt")
    documents_summarize_stuff: str = Field(default="documents/summarize_stuff.pt")
    documents_summarize_refine_qa: str = Field(default="documents/summarize_refine_qa.pt")
    documents_summarize_refine_combine: str = Field(default="documents/summarize_refine_combine.pt")
    documents_summarize_mapreduce_map: str = Field(default="documents/summarize_mapreduce_map.pt")
    documents_summarize_mapreduce_reduce: str = Field(default="documents/summarize_mapreduce_reduce.pt")
    memory_summary: str = Field(default="memory/summary.pt")
    system_out_put_parser: str = Field(default="system/out_put_parser.pt")
    agent_react: str = Field(default="agent/react.pt")
    agent_react_pangu: str = Field(default="agent/react_pangu.pt")
    agent_tool_desc: str = Field(default="agent/tool_desc.pt")
    agent_tool_desc_pangu: str = Field(default="agent/tool_desc_pangu.pt")
    agent_tool_json_schema: str = Field(default="agent/tool_json_schema.pt")
    agent_tool_simple_schema: str = Field(default="agent/tool_simple_schema.pt")
    agent_react_pangu_2: str = Field(default="agent/react_pangu_2.pt")
    agent_tool_desc_pangu_2: str = Field(default="agent/tool_desc_pangu_2.pt")
    agent_react_pangu_unify: str = Field(default="agent/react_pangu_unify.pt")
    conversation_default: str = Field(default="conversation/default.pt")
    skill_conversation_rewrite: str = Field(default="skill/conversation_rewrite.pt")
    skill_agent_session_summary: str = Field(default="skill/agent_session_summary.pt")
    question_long_answer: str = Field(default="question/long_answer.jinja2")
    question_short_answer: str = Field(default="question/short_answer.jinja2")
    question_qa_with_sources_stuff: str = Field(default="question/qa_with_sources_stuff.jinja2")
    question_question_only: str = Field(default="question/question_only.jinja2")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Type

from langchain_pangu.pangukitsappdev.api.doc_split.base import LoaderApi
from langchain_pangu.pangukitsappdev.api.doc_split.split_config import SplitConfig
from langchain_pangu.pangukitsappdev.doc_split.pangu_split import DocPanguSplit


class DocSplits:
    split_map: Dict[str, Type[LoaderApi]] = {}

    @classmethod
    def register(cls, split_type: Type[LoaderApi], split_name: str):
        """
        注册一种split的类型
        :param split_type: split的类型，要求是LoaderApi的子类
        :param split_name: split的名字，唯一代表这个split的名字
        :return: none
        """
        cls.split_map[split_name] = split_type

    @classmethod
    def of(cls, split_name: str, split_config: SplitConfig = None) -> LoaderApi:
        """
        根据名字创建一个LoaderApi的实现类
        :param split_name: split的名字，唯一标识一种分割
        :param split_config: （Optional）split的相关配置，如果不传递则从默认配置文件中或者环境变量中获取
        :return: LoaderApi
        """

        split_type = cls.split_map.get(split_name)
        if not split_type:
            raise ValueError(
                f"Unregistered split name: {split_name}, "
                f"please call register(split_type, split_name) before use.")

        local_split_config = split_config if split_config else cls._load_split_config()

        return split_type(local_split_config)

    @classmethod
    def _load_split_config(cls):
        return SplitConfig()


DocSplits.register(DocPanguSplit, "pangu-doc")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import ABC, abstractmethod


class PythonSandBox(ABC):
    @abstractmethod
    def run(self, code: str) -> str:
        """
        执行一段Python代码
        :param code: Python代码
        :return: Python代码的执行结果
        """


#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
"""
SCC加解密工具类
Usage:
from crypto_util import crypto_util

"""

import logging
import os
import platform
import sys

logger = logging.getLogger(__name__)

SCC_CONF_PATH_ENV = "SCC_CONF_PATH"

SYSTEM_PLATFORM = platform.system().lower()
SCC_LIB_PATH = None
if SYSTEM_PLATFORM == 'windows':
    SCC_LIB_PATH = "C:/Program Files/SecComponent/lib"
elif SYSTEM_PLATFORM == "linux":
    SCC_LIB_PATH = "/usr/local/seccompnent/lib"

if SCC_LIB_PATH:
    sys.path.append(SCC_LIB_PATH)


class SccCryptoAPIWrapper:
    """SCC加解密Api
    需要在机器上安装华为云的scc安全组件后使用这个类。内部封装了CryptoAPI（scc组件安装后获的），在SccCryptoAPIWrapper新创建实例时会对
    CryptoAPI做初始化。

    Attributes:
        scc_conf_path: scc的配置路径
        _scc_api: 懒加载的CryptoAPI类实例，初始化的时候是None
    """

    def __init__(self, scc_conf_path=None):
        """
        Args:
            scc_conf_path: （Option）scc.conf的路径，不传递则从环境变量SCC_CONF_PATH获取
        """
        self._scc_api = None
        self.scc_conf_path = scc_conf_path

    def __del__(self):
        if self._scc_api:
            self._scc_api.finalize()

    def _init_scc(self):
        if self._scc_api:
            return

        conf_path = self.scc_conf_path if self.scc_conf_path else os.getenv(SCC_CONF_PATH_ENV)
        if not conf_path or not os.path.exists(conf_path):
            raise ValueError(
                "Please input scc_conf_path as params when __init__ "
                "or set env SCC_CONF_PATH and the conf file should exist!")

        try:
            from CryptoAPI import CryptoAPI
        except ImportError:
            raise ValueError(
                f"Can't import CryptoAPI, "
                f"please check scc has been installed and the {SYSTEM_PLATFORM}: {SCC_LIB_PATH} exists ")

        scc_api = CryptoAPI()
        scc_api.initialize(conf_path)
        if scc_api.gInitFlag:
            logger.info("SCC init success")
            self._scc_api = scc_api
        else:
            logger.error("SCC init failed, please check the logfile config in scc.conf")

    def encrypt(self, plain_text: str) -> str:
        """
        加密
        :param plain_text: 明文敏感信息
        :return: 加密后的密文
        """
        self._init_scc()
        return self._scc_api.encrypt(plain_text)

    def decrypt(self, cipher: str) -> str:
        """
        解密
        :param cipher: 密文
        :return: 解密后的明文
        """
        self._init_scc()
        return self._scc_api.decrypt(cipher)


crypto_util = SccCryptoAPIWrapper()


def decrypt(key_id: str, cipher: str) -> str:
    return crypto_util.decrypt(cipher)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import ABC, abstractmethod
from typing import Dict, List

from langchain.embeddings.base import Embeddings

from langchain_pangu.pangukitsappdev.api.embeddings.embedding_config import EmbeddingConfig


class EmbeddingApi(Embeddings, ABC):
    @abstractmethod
    def embed_qa_documents(self, doc_texts: List[Dict[str, str]], weight: Dict[str, int]) -> List[List[float]]:
        """
        支持对一份数据的不同文本分权重进行Embedding
        :param doc_texts: 批量索引的文本
        :param weight: 权重，key值和doc_text的元素中的key值一致
        :return: embedding结果
        """
        pass


class AbstractEmbeddingApi(EmbeddingApi, ABC):

    def __init__(self, embedding_config: EmbeddingConfig):
        self.embedding_config = embedding_config
        self.embeddings = self.create_embeddings(embedding_config)

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        return self.embeddings.embed_documents(texts)

    def embed_query(self, text: str) -> List[float]:
        return self.embeddings.embed_query(text)

    def embed_qa_documents(self, doc_texts: List[Dict[str, str]], weight: Dict[str, int]) -> List[List[float]]:
        # 默认未实现，直接抛异常
        raise NotImplementedError("Unsupported operation!")

    @abstractmethod
    def create_embeddings(self, embedding_config: EmbeddingConfig) -> Embeddings:
        """
        创建embeddings的实现类
        :param embedding_config: 配置
        :return: embeddings的实现类
        """
        pass



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Any

from pydantic.v1 import BaseModel, Field


class BulkData(BaseModel):
    """
    ES Bulk消息
    Attributes:
        id: 索引ID
        data: 数据
    """
    id: str
    data: Dict[str, Any] = Field(default={})



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import ABC, abstractmethod
from typing import List

from langchain_pangu.pangukitsappdev.api.memory.vector.base import Document


class WebSearch(ABC):
    @abstractmethod
    def query(self, query: str, top_k: int) -> List[Document]:
        """
        执行query
        :param query: 查询体
        :param top_k: 查询记录数
        :return: 相关文档
        """


#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import ABC, abstractmethod
from inspect import signature, Signature
from typing import Optional, Any, Callable, Type

from langchain.tools import BaseTool
from pydantic.v1 import BaseModel

DEFAULT_SINGLE_ARG = "arg"


class AbstractTool(BaseTool, ABC):
    input_desc: str
    output_desc: str
    principle: str
    func: Optional[Callable[..., str]]
    return_type: Optional[Type]

    def get_tool_id(self):
        """
        工具的唯一标识，在同一个Agent里必须唯一
        Returns: str, 工具的标识
        """
        return self.name

    def get_tool_desc(self) -> str:
        """
        工具的描述，唯一值
        Returns: str, 工具的描述
        """
        return self.description

    @property
    def input_type(self):
        """
        工具的input类型，
        Returns: Type, 工具input类型
        """
        if self.args_schema is not None:
            return self.args_schema
        else:
            function = self.func if self.func else self._run
            params = list(signature(function).parameters.values())
            if len(params) != 1 and len(params) != 0:
                raise ValueError("Input param should extend BaseTool or have less than one simple type")
            return params[0].annotation if params else type(None)

    @property
    def output_type(self):
        """
        工具的返回值类型，此方法可获取func或_run方法返回类型
        Returns: Type, 工具的返回值类型
        """
        if self.func:
            # 动态工具获取self.func类型
            func_type = signature(self.func).return_annotation
        else:
            # 静态工具重写self._run方法指定类型
            func_type = signature(self._run).return_annotation
        # 若为lambda表达式或未指定返回类型函数，需要创建时指定return_type
        self.return_type = self.return_type if func_type == Signature.empty else func_type
        if not self.return_type:
            raise ValueError("Dynamic func method or static _run method must specify return annotation!")
        return self.return_type

    @abstractmethod
    def get_input_schema(self) -> str:
        """
        获取入参的Json Schema定义
        Returns: str, 获取入参的Json Schema定义
        """

    @abstractmethod
    def get_output_schema(self) -> str:
        """
        获取出参的Json Schema定义
        Returns: str, 获取出参的Json Schema定义
        """

    @abstractmethod
    def get_pangu_function(self) -> str:
        """
        获取pangu Function的定义
        Returns: str, pangu Function的定义
        """


class PanguFunction(BaseModel):
    """PanguFunction参数
    Attributes:
        name: tool名称
        description: tool功能描述，描述tool的作用
        arguments: tool输入
        principle: tool使用原则，告诉模型在什么情况下使用tool
        results: tool输出
    """
    name: Optional[str]
    description: Optional[str]
    arguments: Optional[Any]
    principle: Optional[str]
    results: Optional[Any]



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Any

from langchain.cache import RedisCache, RedisSemanticCache
from langchain.embeddings.base import Embeddings
from langchain.schema.cache import RETURN_VAL_TYPE
from redis import Redis

from langchain_pangu.pangukitsappdev.api.memory.cache.base import CacheApiAdapter
from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import CacheStoreConfig


class RedisCacheApi(CacheApiAdapter):
    """支持TTL的缓存策略，内部封装Langchain的redis缓存
    1. 通过CacheStoreConfig创建Redis实例，从而初始化RedisCache\n
    """

    def __init__(self, cache_config: CacheStoreConfig):
        if cache_config.expire_after_access > 0:
            raise ValueError("Redis Cache do not support expire_after_access")
        if cache_config.maximum_size > 0:
            raise ValueError("Redis Cache do not support maximum_size")
        redis_client: Redis = Redis.from_url(cache_config.server_info.get_urls()[0])

        redis_cache = TTLRedisCache(redis_client,
                                    cache_config.expire_after_write)
        super().__init__(redis_cache, cache_config.session_tag)


class TTLRedisCache(RedisCache):
    """支持TTL的缓存策略，继承自langchain的BaseCache
    """

    def __init__(self, redis_client, ttl):
        super().__init__(redis_=redis_client)
        self.expire_after_write = ttl

    def update(self, prompt: str, llm_string: str, return_val: RETURN_VAL_TYPE) -> None:
        super().update(prompt, llm_string, return_val)
        if self.expire_after_write > 0:
            self.redis.expire(name=self._key(prompt, llm_string), time=self.expire_after_write)

    """重写clear方法，统一CacheApiAdapter入参接口
    """
    def clear(self, **kwargs: Any):
        super().clear()


class RedisSemanticCacheApi(CacheApiAdapter):
    """Redis的语义缓存
    内部封装pangukitsappdev.cache.redis_ext.TTLRedisSemanticCache。支持配置缓存的ttl
    """

    def __init__(self, cache_config: CacheStoreConfig):
        config = {
            "redis_url": cache_config.server_info.get_urls()[0],
            "embedding": cache_config.embedding,
            "score_threshold": cache_config.score_threshold,
        }
        # 目前langchain不支持RedisSemanticCache，ttl功能与langchain同步，无需锁定版本
        if "ttl" in RedisSemanticCache.__init__.__code__.co_varnames and cache_config.expire_after_write > 0:
            config.update({"ttl": cache_config.expire_after_write})

        redis_semantic_cache = RedisSemanticCache(**config)
        super().__init__(redis_semantic_cache, cache_config.session_tag)


class TTLRedisSemanticCache(RedisSemanticCache):
    def __init__(self, redis_url: str,
                 embedding: Embeddings,
                 ttl,
                 score_threshold: float = 0.2,
                 ):
        super().__init__(redis_url, embedding, score_threshold)
        self.ttl = ttl

    def update(self, prompt: str, llm_string: str, return_val: RETURN_VAL_TYPE) -> None:
        """更新缓存，并配置缓存的ttl
        逻辑复用父类逻辑，获取ids进行ttl配置
        Args:
            prompt: prompt query
            llm_string: 唯一标识llm
            return_val: llm返回的数据
        """
        llm_cache = self._get_llm_cache(llm_string)
        # Write to vectorstore
        metadata = {
            "llm_string": llm_string,
            "prompt": prompt,
            "return_val": [generation.text for generation in return_val],
        }
        keys = llm_cache.add_texts(texts=[prompt], metadatas=[metadata])
        llm_cache.client.expire(name=keys[0], time=self.ttl)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Type

from langchain_pangu.pangukitsappdev.api.common_config import IAMConfigWrapper, HttpConfig
from langchain_pangu.pangukitsappdev.api.embeddings.base import EmbeddingApi
from langchain_pangu.pangukitsappdev.api.embeddings.embedding_config import EmbeddingConfig
from langchain_pangu.pangukitsappdev.embeddings.openai import OpenAIEmbeddingApi
from langchain_pangu.pangukitsappdev.embeddings.pangu import CSSEmbeddingApi
from langchain_pangu.pangukitsappdev.embeddings.tool import ToolEmbeddingApi


class Embeddings:
    embeddings_map: Dict[str, Type[EmbeddingApi]] = {}

    @classmethod
    def register(cls, embedding_type: Type[EmbeddingApi], embedding_name: str):
        """
        注册一种embedding的类型
        :param embedding_type: embedding的类型，要求是EmbeddingApi的子类
        :param embedding_name: embedding的名字，唯一代表这个embedding的名字
        :return: none
        """
        cls.embeddings_map[embedding_name] = embedding_type

    @classmethod
    def of(cls, embedding_name: str, embedding_config: EmbeddingConfig = None) -> EmbeddingApi:
        """
        根据名字创建一个EmbeddingApi的实现类
        :param embedding_name: embedding的名字，唯一标识一种Embedding
        :param embedding_config: （Optional）Embedding的相关配置，如果不传递则从默认配置文件中或者环境变量中获取
        :return: EmbeddingApi
        """

        embedding_type = cls.embeddings_map.get(embedding_name)
        if not embedding_type:
            raise ValueError(
                f"Unregistered embedding name: {embedding_name}, "
                f"please call register(embedding_type, embedding_name) before use.")

        local_embedding_config = embedding_config if embedding_config else cls._load_embedding_config(embedding_name)

        return embedding_type(local_embedding_config)

    @classmethod
    def _load_embedding_config(cls, embedding_name: str):
        if embedding_name == "tool":
            return EmbeddingConfig(env_prefix="sdk.embedding.tool",
                                   iam_config=IAMConfigWrapper(env_prefix="sdk.embedding.tool.iam").get_iam_config(),
                                   http_config=HttpConfig(env_prefix="sdk.embedding.tool.proxy"))
        else:
            return EmbeddingConfig()


Embeddings.register(CSSEmbeddingApi, "css")
Embeddings.register(OpenAIEmbeddingApi, "openAI")
Embeddings.register(ToolEmbeddingApi, "tool")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import List, Dict

import requests
from langchain.embeddings.base import Embeddings
from pydantic.v1 import BaseModel, Extra

from langchain_pangu.pangukitsappdev.api.common_config import AUTH_TOKEN_HEADER
from langchain_pangu.pangukitsappdev.api.embeddings.base import AbstractEmbeddingApi
from langchain_pangu.pangukitsappdev.api.embeddings.embedding_config import EmbeddingConfig
from langchain_pangu.pangukitsappdev.auth.iam import IAMTokenProvider, IAMTokenProviderFactory


def _default_doc_map(text: str) -> dict:
    """默认文档拆分方法
    封装成{"content": text}返回
    :param text: 文本内容
    :return: :class:`dict` {"content": text}
    :rtype: dict
    """
    return {"content": text}


def json_doc_map(text: str, encoding: str = "utf8") -> Dict[str, str]:
    import json
    return json.loads(text, encoding=encoding)


class WeightPanguEmbeddings(Embeddings, BaseModel):
    """支持文本拆分，不同部分使用不同的权重做Embedding

    """

    url: str
    token_provider: IAMTokenProvider
    proxies: dict = {}

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore
        arbitrary_types_allowed = True

    def auth_headers(self) -> dict:

        token = self.token_provider.get_valid_token()
        headers = {
            AUTH_TOKEN_HEADER: token,
            "X-Agent": "pangu-kits-app-dev"
        } if token else {"X-Agent": "pangu-kits-app-dev"}
        return headers

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """批量对文本做Embedding
        一般用在构造向量数据库的时候批量进行文档Embedding

        :param texts: 文档数据
        :return: :class:`list[list[float]] <list[list[float]]>` 以二维数组格式返回的向量数据
        :rtype: list[list[float]]
        """

        request_body = {
            "query": texts
        }

        headers = self.auth_headers()
        batch_url = f"{self.url}/embedding/query"
        rsp = requests.post(batch_url, headers=headers, json=request_body, proxies=self.proxies, verify=False)

        if 200 == rsp.status_code:
            result = rsp.json()["embedding"]
            return result

        rsp.raise_for_status()

    def embed_qa_documents(self, doc_texts: List[Dict[str, str]], weight: Dict[str, int]) -> List[List[float]]:
        request_body = {
            "weight": weight,
            "docs": doc_texts
        }

        headers = self.auth_headers()

        batch_url = f"{self.url}/embedding/batch"
        rsp = requests.post(batch_url, headers=headers, json=request_body, proxies=self.proxies, verify=False)

        if 200 == rsp.status_code:
            result = rsp.json()["embedding"]
            return result

        rsp.raise_for_status()

    def embed_query(self, text: str) -> List[float]:
        """对query词做Embedding

        检索时对查询词做Embedding

        :param text: 查询词
        :return: :class:`List[float] <List[float]>` 向量数据
        :rtype: List[float]
        """
        return self.embed_documents([text])[0]


    def embed_split_documents(self, documents: List, weight: Dict[str, int]) -> List[List[float]]:
        """
        根据weight指定需要embedding的字段及其权重
        :param documents: 文档列表
        :param weight: 字段权重
        :return:
        """
        doc_texts = []
        # 需要判断field是否是document的属性
        if not weight.keys() <= documents[0].dict().keys():
            raise Exception("weight keys should be in documents")

        for document in documents:
            doc_text = {}
            for field in weight:
                doc_text[field] = getattr(document, field)
            doc_texts.append(doc_text)

        return self.embed_qa_documents(doc_texts, weight)


class CSSEmbeddingApi(AbstractEmbeddingApi):
    """
    CSS的Embedding API，支持embed_qa_documents
    """

    def embed_qa_documents(self, doc_texts: List[Dict[str, str]], weight: Dict[str, int]) -> List[List[float]]:
        return self.embeddings.embed_qa_documents(doc_texts=doc_texts, weight=weight)

    def create_embeddings(self, embedding_config: EmbeddingConfig) -> Embeddings:
        token_provider = IAMTokenProviderFactory.create(embedding_config.iam_config)
        return WeightPanguEmbeddings(url=embedding_config.css_url,
                                     token_provider=token_provider,
                                     proxies=embedding_config.http_config.requests_proxies())

    def embed_split_documents(self, doc_texts: List, weight: Dict[str, int]) -> List[List[float]]:
        return self.embeddings.embed_split_documents(doc_texts, weight)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from elasticsearch.connection.http_requests import RequestsHttpConnection


class ProxyRequestsHttpConnection(RequestsHttpConnection):
    def __init__(
            self,
            proxies=None,
            **kwargs
    ):
        """
        初始化
        Args:
            proxies: (Optional) 包含 proxies 参数，配置 es 代理信息
        """
        super(ProxyRequestsHttpConnection, self).__init__(
            **kwargs
        )
        if proxies:
            self.session.proxies = proxies



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Any, Optional

from langchain.cache import SQLAlchemyCache
from langchain.schema import Generation
from langchain.schema.cache import RETURN_VAL_TYPE
from sqlalchemy import Column, Integer, Text, String, create_engine, Engine, select, delete, make_url
from sqlalchemy.orm import declarative_base, Session

from langchain_pangu.pangukitsappdev.api.memory.cache.base import CacheApiAdapter
from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import CacheStoreConfig
from langchain_pangu.pangukitsappdev.utils.time_date import now_yyyyMMddHHmmss, now_sec, to_yyyyMMddHHmmss

PROMPT_PREFIX_LEN = 250

Base = declarative_base()

SQL_CACHE_TABLE_NAME = "tbl_full_llm_cache_py"


class MySQLLLMCacheSchema(Base):
    """MySQL table for full LLM Cache (all generations)."""

    __tablename__ = SQL_CACHE_TABLE_NAME
    id = Column(Integer, primary_key=True, autoincrement=True)
    prompt = Column(Text())
    prompt_prefix = Column(String(765), nullable=True, default=None, index=True)
    session_tag = Column(String(511), nullable=True, default=None, index=True)
    answer = Column(Text(), nullable=True)
    create_date = Column(String(45), nullable=True, default=None, index=True)


SCHEMA_LOOKUP = {
    "mysql": MySQLLLMCacheSchema
}


class SqlCacheApi(CacheApiAdapter):

    def __init__(self, cache_config: CacheStoreConfig):
        if cache_config.expire_after_access > 0:
            raise ValueError("Sql Cache do not support expire_after_access")
        url_instance = make_url(cache_config.server_info.get_urls()[0])
        backend_engine = url_instance.get_backend_name()
        """默认使用mysql"""
        schema_clz = SCHEMA_LOOKUP.get(backend_engine, MySQLLLMCacheSchema)
        sql_alchemy_cache = SQLCache(engine=create_engine(url_instance),
                                     cache_schema=schema_clz,
                                     ttl=cache_config.expire_after_write,
                                     maximum_size=cache_config.maximum_size,
                                     pool_size=cache_config.server_info.pool_size)
        super().__init__(sql_alchemy_cache, cache_config.session_tag)


class SQLCache(SQLAlchemyCache):

    def __init__(self, engine: Engine, cache_schema, ttl, maximum_size, pool_size=5):
        super().__init__(engine, cache_schema)
        self.ttl = ttl
        self.maximum_size = maximum_size if maximum_size > 0 else float('inf')

    def lookup(self, prompt: str, llm_string: str) -> Optional[RETURN_VAL_TYPE]:
        self.evict(prompt, llm_string)

        stmt = (
            select(self.cache_schema.answer)
            .where(self.cache_schema.session_tag == llm_string)
            .where(self.cache_schema.prompt_prefix == prompt[:PROMPT_PREFIX_LEN])
            .where(self.cache_schema.prompt == prompt)
        )
        with Session(self.engine) as session:
            rows = session.execute(stmt).fetchall()
            if rows:
                return [Generation(text=row[0]) for row in rows]
        return None

    def update(self, prompt: str, llm_string: str, return_val: RETURN_VAL_TYPE) -> None:
        items = [
            self.cache_schema(prompt=prompt,
                              prompt_prefix=prompt[:PROMPT_PREFIX_LEN],
                              session_tag=llm_string, answer=gen.text, create_date=now_yyyyMMddHHmmss())
            for i, gen in enumerate(return_val)
        ]
        with Session(self.engine) as session, session.begin():
            for item in items:
                session.merge(item)
        self.delete_by_batch(llm_string)

    def clear(self, **kwargs: Any) -> None:
        session_tag = kwargs.get("llm_string")

        if session_tag:
            with Session(self.engine) as session:
                session.query(self.cache_schema).where(self.cache_schema.session_tag == session_tag).delete()
                session.commit()
            return

        with Session(self.engine) as session:
            session.query(self.cache_schema).delete()
            session.commit()

    def evict(self, prompt, session_tag):
        if self.ttl <= 0:
            return
        del_stmt = (
            delete(self.cache_schema)
            .where(self.cache_schema.session_tag == session_tag)
            .where(self.cache_schema.prompt_prefix == prompt[:PROMPT_PREFIX_LEN])
            .where(self.cache_schema.create_date < to_yyyyMMddHHmmss(now_sec() - self.ttl))
            .where(self.cache_schema.prompt == prompt)
        )

        with Session(self.engine) as session:
            session.execute(del_stmt)
            session.commit()

    def delete_by_batch(self, llm_string: str):
        with Session(self.engine) as session:
            records = session.query(self.cache_schema)\
                .where(self.cache_schema.session_tag == llm_string).order_by(self.cache_schema.id).all()
            count = len(records)
            if count > self.maximum_size:
                for i in range(count - self.maximum_size):
                    session.delete(records[i])
                session.commit()




#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional, List

from pydantic.v1 import Field

from langchain_pangu.pangukitsappdev.api.common_config import ServerInfo, HttpConfig
from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings
from langchain_pangu.pangukitsappdev.api.embeddings.base import EmbeddingApi


class ServerInfoCss(ServerInfo):

    def get_urls(self) -> [str]:
        return self.get_http_urls()


class VectorStoreConfig(SdkBaseSettings):
    """
    向量库相关配置
    Attributes:
        store_name: 存储名称
        server_info: ServerInfoCss，存储服务器链接相关信息，默认使用sdk.store.css前缀的配置
        index_name: 索引名称
        embedding: embedding库
        distance_strategy: 检索模式，默认使用内积inner_product
        text_key: 文本内容字段名
        metadata_key: 扩展元数据字段名
        bulk_size: 文本入库时， 每批次记录数； 默认50， 受向量库及模型能力影响
        vector_fields: CSS向量数据中，如果启用了embedding插件，用来指定那些字段做向量化查询
        source_fields: 指定那些字段作为source字段返回（ES）,不配置默认全部取回
        ttl: 老化时间，单位 秒；向量存储场景，默认不设置TTL
        verify_certs: 是否校验https的证书。默认是True
        http_config: http代理配置
    """

    store_name: Optional[str]
    server_info: ServerInfo = Field(default_factory=lambda: ServerInfoCss(env_prefix="sdk.memory.css"))
    index_name: Optional[str]
    embedding: Optional[EmbeddingApi]
    distance_strategy: str = Field(default="inner_product")
    text_key: str = Field(default="text")
    metadata_key: str = Field(default="metadata")
    bulk_size: int = Field(default=50)
    vector_fields: List[str] = Field(default=[])
    source_fields: List[str] = Field(default=[])
    ttl: Optional[int]
    verify_certs: bool = Field(default=True)
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.memory.css.proxy"))



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.

try:
    from langchain.vectorstores import VectorStore
except ImportError:
    from langchain.schema.vectorstore import VectorStore

from langchain_pangu.pangukitsappdev.api.memory.vector.base import AbstractVectorApi
from langchain_pangu.pangukitsappdev.api.memory.vector.vector_config import VectorStoreConfig
from langchain_pangu.pangukitsappdev.vectorstores.css_store import CSSVectorSearch


class CSSVectorApi(AbstractVectorApi):
    def create_vector_store(self, vector_config: VectorStoreConfig) -> VectorStore:
        config = {
            "elasticsearch_url": vector_config.server_info.get_urls(),
            "index_name": vector_config.index_name,
            "embedding": vector_config.embedding,
            "verify_certs": vector_config.verify_certs,
            "text_key": vector_config.text_key,
            "vector_fields": vector_config.vector_fields,
            "source_fields": vector_config.source_fields,
            "proxies": vector_config.http_config.requests_proxies()
        }

        return CSSVectorSearch(**config)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.

import time
from typing import List

import requests

from langchain_pangu.pangukitsappdev.api.common_config import AUTH_TOKEN_HEADER
from langchain_pangu.pangukitsappdev.api.doc_split.base import AbstractLoaderApi
from langchain_pangu.pangukitsappdev.api.doc_split.split_config import SplitConfig
from langchain_pangu.pangukitsappdev.api.memory.vector.base import Document
from langchain_pangu.pangukitsappdev.auth.iam import IAMTokenProviderFactory


def extract_actions(result):
    paragraphs = []
    for page in result['pages']:
        page_num = page['page_num']
        for component in page['components']:
            paragraphs.append(
                Document(
                    page_content=component.get("text", ""),
                    metadata={
                        "id": component['id'],
                        "title": component.get("title", ""),
                        "component_num": component["component_num"],
                        "page_num": page_num,
                        "url": result['doc_name']
                    }))
    return paragraphs


class DocPanguSplit(AbstractLoaderApi):

    def __init__(self, split_config: SplitConfig):
        super().__init__(split_config)
        self.token_provider = IAMTokenProviderFactory.create(self.split_config.iam_config)
        self.proxies = self.split_config.http_config.requests_proxies()

    """
    文档解析，目前支持doc/pdf
    """

    def load(self) -> List[Document]:
        token = self.token_provider.get_valid_token()

        headers = {
            AUTH_TOKEN_HEADER: token
        } if token else {}

        task_id = self.submit_task(headers, self.split_config.mode)
        result = self.get_task_result(task_id, headers)
        docs = extract_actions(result)
        return docs

    def submit_task(self, headers, mode=0) -> str:
        """
        提交文档解析任务
        :param headers: 鉴权信息
        :param mode: 段落拆分模式，默认为0
        0 - 返回文档的原始段落，不做其他处理
        1 - 根据标注的书签或目录分段，一般适合有层级标签的word文档
        2 - 根据内容里的章节条分段，适合制度类文档
        3 - 根据长度分段，默认按照500字拆分，会尽量保留完整句子
        :return:
        """
        if mode not in [0, 1, 2, 3]:
            raise Exception('Unsupported mode value.')
        payload = {'mode': mode}
        files = [
            ('file', (self.split_config.file_name, open(self.split_config.file_path, 'rb')))
        ]
        res = requests.post(self.split_config.upload_url(), headers=headers,
                            data=payload, files=files, proxies=self.proxies)
        if res.status_code == 200:
            res = res.json()
            if 'task_id' in res:
                return res['task_id']
        raise Exception(f'upload file failed, error code is {res["error_code"]}, reason is {res["error_msg"]}')

    def get_task_result(self, task_id, headers):
        res = requests.get(self.split_config.result_url(task_id), headers=headers, proxies=self.proxies).json()
        while res['task_status'] not in ['SUCCESS', 'ERROR']:
            time.sleep(3)
            res = requests.get(self.split_config.result_url(task_id), headers=headers, proxies=self.proxies).json()
        if res['task_status'] == 'SUCCESS':
            return res['result']
        raise Exception(f'task result failed, error code is {res["error_code"]}, reason is {res["error_msg"]}')



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Type, Dict

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.api.llms.llm_config import LLMConfig, LLMModuleConfig
from langchain_pangu.pangukitsappdev.llms.gallery import GalleryLLMApi
from langchain_pangu.pangukitsappdev.llms.openai import OpenAILLMApi
from langchain_pangu.pangukitsappdev.llms.pangu import PanguLLMApi


class LLMs:
    llms_map: Dict[str, Type] = {}

    @classmethod
    def register(cls, llm_type: Type[LLMApi], llm_name: str):
        """
        注册一种llm的类型
        :param llm_type: llm的类型，要求是BaseLLM的子类
        :param llm_name: llm的名字，唯一代表这个llm的名字
        :return: none
        """
        cls.llms_map[llm_name] = llm_type

    @classmethod
    def of(cls, llm_name: str, llm_config: LLMConfig = None) -> LLMApi:
        """
        根据名字创建一个LLMApi的实现类
        :param llm_name: llm的名字，唯一标识一种LLM
        :param llm_config: （Optional）LLM的相关配置，如果不传递则从默认配置文件中或者环境变量中获取
        :return: LLMApi
        """

        llm_type = cls.llms_map.get(llm_name)
        if not llm_type:
            raise ValueError(f"Unregistered llm name: {llm_name}, please call register(llm_type, llm_name) before use.")

        local_llm_config = llm_config if llm_config else cls._load_llm_config()

        return llm_type(local_llm_config)

    @classmethod
    def of_module(cls, llm_name: str, llm_module_config: LLMModuleConfig) -> LLMApi:
        """根据指定的LLMModuleConfig构造LLMApi
        Args:
            llm_name:  llm的名字，唯一标识一种LLM
            llm_module_config: 外部参数传递的LLMModuleConfig配置

        Returns: LLMApi

        """
        llm_config = LLMConfig(llm_module_config=llm_module_config)
        return cls.of(llm_name, llm_config)

    @classmethod
    def of_env_prefix(cls, llm_name: str, env_prefix) -> LLMApi:
        """

        Args:
            llm_name: llm的名字，唯一标识一种LLM
            env_prefix: 环境变量或者配置key的前缀

        Returns: LLMApi:

        """
        llm_module_config = LLMModuleConfig(env_prefix=env_prefix)
        return cls.of_module(llm_name, llm_module_config)

    @classmethod
    def _load_llm_config(cls):
        return LLMConfig()


LLMs.register(PanguLLMApi, "pangu")
LLMs.register(OpenAILLMApi, "openAI")
LLMs.register(GalleryLLMApi, "gallery")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import ABC, abstractmethod
from typing import Optional

from langchain.schema import Generation
from langchain.schema.cache import BaseCache

from langchain_pangu.pangukitsappdev.api.schema import LLMResp


class CacheApi(ABC):
    """SDK封装的缓存对外接口"""

    @abstractmethod
    def lookup(self, prompt: str) -> LLMResp:
        """查询

        Args:
            prompt: 提示词

        Returns: 查询结果

        """

    @abstractmethod
    def update(self, prompt: str, value: LLMResp) -> None:
        """
        更新
        Args:
            prompt: 提示词
            value: 回答结果

        Returns:
            LLMResp
        """

    @abstractmethod
    def clear(self) -> None:
        """
        清空
        Args:

        Returns:
            LLMResp
        """


class CacheApiAdapter(CacheApi):
    """对Langchain原生的缓存做适配

    Attributes:
        cache: BaseCache的实现类，封装的langchain的cache
    """

    def __init__(self, cache: BaseCache, session_tag: str = ""):
        self.cache = cache
        self.session_tag = session_tag

    def lookup(self, prompt: str) -> Optional[LLMResp]:
        cached_llm_result = self.cache.lookup(prompt=prompt, llm_string=self.session_tag)
        if isinstance(cached_llm_result, list) and cached_llm_result:
            answer = cached_llm_result[0].text
            return LLMResp(answer=answer, is_from_cache=True)
        else:
            return None

    def update(self, prompt: str, value: LLMResp) -> None:
        self.cache.update(prompt=prompt,
                          llm_string=self.session_tag,
                          return_val=[Generation(text=value.answer)])

    def clear(self) -> None:
        self.cache.clear(llm_string=self.session_tag)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import ABC, abstractmethod
from typing import Any, Dict, Union, List

from langchain.chains.llm import LLMChain
from langchain.schema.prompt_template import BasePromptTemplate

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi, ConversationMessage
from langchain_pangu.pangukitsappdev.api.llms.llm_config import LLMParamConfig


class Skill(ABC):

    @abstractmethod
    def execute(self, inputs: Dict[str, Any]) -> str:
        """
        执行Executor
        :param inputs: 输入的参数，基本上都是用来渲染prompt_template的
        :return: 执行结果，LLM的返回结果
        """
        pass

    @staticmethod
    def skill_llm_ask(prompt: Union[str, List[ConversationMessage]],
                      llm_api: LLMApi,
                      param_config: LLMParamConfig = None) -> str:
        """
        支持skill流式输出
        :param prompt: 用户输入
        :param llm_api: 模型api接口
        :param param_config: 可选参数
        :return: 执行结果，LLM的返回结果
        """
        if (param_config and param_config.stream) or llm_api.get_llm_config().llm_param_config.stream:
            tokens = llm_api.ask(prompt, param_config)
            answer = ""
            for token in tokens:
                answer += token
        else:
            answer = llm_api.ask(prompt, param_config).answer
        return answer


class SimpleSkill(Skill):
    """
    一个Executor的简单实现，传递prompt_template和llm_api两个参数即可，面向api包下面的接口编程
    """

    def __init__(self, prompt_template: BasePromptTemplate, llm_api: LLMApi):
        self.prompt_template = prompt_template
        self.llm_api = llm_api

    def execute(self, inputs: Dict[str, Any], param_config: LLMParamConfig = None) -> str:
        prompt = self.prompt_template.format(**inputs)
        return self.skill_llm_ask(prompt, self.llm_api, param_config)


class ChainWrappedSkill(Skill):
    """
    通过封装一个Chain来实现Executor，方便集成langchain预置的一些chain
    """

    def __init__(self, chain: LLMChain):
        self.chain = chain

    def execute(self, inputs: Dict[str, Any]) -> str:
        # 默认返回output_keys的第一个元素所代表的输出内容，一般都是回答
        chain_output_key = self.chain.output_keys[0]

        return self.chain(inputs)[chain_output_key]



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Type

from langchain_pangu.pangukitsappdev.api.memory.cache.base import CacheApi
from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import CacheStoreConfig, ServerInfoSql
from langchain_pangu.pangukitsappdev.cache.gptcache_adapter import SemanticGptCacheApi
from langchain_pangu.pangukitsappdev.cache.in_memory import InMemoryCacheApi
from langchain_pangu.pangukitsappdev.cache.sql import SqlCacheApi
from langchain_pangu.pangukitsappdev.cache.ttl_redis import RedisCacheApi, RedisSemanticCacheApi


class Caches:
    caches_map: Dict[str, Type[CacheApi]] = {}

    @classmethod
    def register(cls, cache_type: Type[CacheApi], cache_name: str):
        """注册一种cache的类型
        Args:
            cache_type: cache的类型，要求是CacheApi的子类
            cache_name: cache的名字，唯一代表这个cache的名字
        """
        cls.caches_map[cache_name] = cache_type

    @classmethod
    def of(cls, cache_name: str, cache_config: CacheStoreConfig = None) -> CacheApi:
        """根据名字创建一个CacheApi的实现类
        cache_name和cache_config.store_name不可同时为空
        Args:
            cache_name: （Optional）cache的名字，唯一标识一个CacheApi的实现类
            cache_config: （Optional）CacheApi的相关配置，如果不传递则从默认配置文件中或者环境变量中获取

        Returns:
            一个CacheApi的实现类

        """
        local_cache_name = cache_name if cache_name else cache_config.store_name

        if not local_cache_name:
            raise ValueError("The parameter cache_name or cache_config.store_name should have value")

        cache_type = cls.caches_map.get(local_cache_name)
        if not cache_type:
            raise ValueError(
                f"Unregistered cache name: {local_cache_name}, \
                    please call register(cache_type, cache_name) before use.")

        local_cache_config = cache_config if cache_config else cls._load_config(local_cache_name)
        if "sql" in local_cache_name and not isinstance(local_cache_config.server_info, ServerInfoSql):
            local_cache_config.server_info = ServerInfoSql(env_prefix="sdk.memory.rds")

        return cache_type(local_cache_config)

    @classmethod
    def _load_config(cls, local_cache_name):
        if "sql" in local_cache_name:
            server_info = ServerInfoSql(env_prefix="sdk.memory.rds")
            return CacheStoreConfig(server_info=server_info)

        return CacheStoreConfig()


Caches.register(RedisCacheApi, "redis")
Caches.register(RedisSemanticCacheApi, "semantic_redis")
Caches.register(InMemoryCacheApi, "inMemory")
Caches.register(SemanticGptCacheApi, "semantic_gptcache")
Caches.register(SqlCacheApi, "sql")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.

from langchain.prompts import PromptTemplate

CN_DEFAULT_REFINE_PROMPT_TMPL = """
【新给出的上下文信息】：
------------
{context_str}
------------
【给定问题】：
{question}
【已有答案】：
{existing_answer}

请再结合【新给出的上下文信息】和【已有答案】，继续回答【给定问题】。
请严格按照如下要求进行回答：
1. 如果【新给出的上下文信息】没有帮助，则不做任何修改直接返回【已有答案】的内容。
2. 请直接返回最终答案，不要返回回答的过程。
3. 答案内容中删除“根据”关键字。
4. 如果根据给出的信息你无法回答，就说不知道，不要编造答案。
"""

CN_DEFAULT_REFINE_PROMPT = PromptTemplate(
    input_variables=["question", "existing_answer", "context_str"],
    template=CN_DEFAULT_REFINE_PROMPT_TMPL,
)

CN_DEFAULT_TEXT_QA_PROMPT_TMPL = (
    "上下文信息如下 \n"
    "---------------------\n"
    "{context_str}"
    "\n---------------------\n"
    "没有任何先验信息，请根据给出的上下文信息回答【给定问题】：{question}\n"
    "如果根据给出的信息你无法回答，就说不知道，不要编造答案。"
)

CN_DEFAULT_TEXT_QA_PROMPT = PromptTemplate(
    input_variables=["context_str", "question"], template=CN_DEFAULT_TEXT_QA_PROMPT_TMPL
)


#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import List

from langchain_pangu.pangukitsappdev.api.tool.base import AbstractTool
from langchain_pangu.pangukitsappdev.retriever.retrieved_tool import RetrievedTool
from langchain_pangu.pangukitsappdev.tool.tool_provider import ToolProvider


class InMemoryToolProvider(ToolProvider):
    """工具类持久化(内存)
    """
    def __init__(self):
        self.tool_store: dict[str, AbstractTool] = {}

    def provide(self, retrieved_tools: List[RetrievedTool], query: str) -> List[AbstractTool]:
        tools = []
        for retrieved_tool in retrieved_tools:
            if retrieved_tool.tool_id in self.tool_store.keys():
                tools.append(self.tool_store.get(retrieved_tool.tool_id))
        return tools

    def add(self, tools: List[AbstractTool]) -> List[str]:
        tool_ids = []
        for tool in tools:
            tool_ids.append(tool.get_tool_id())
            self.tool_store.update({tool.get_tool_id(): tool})
        return tool_ids

    def remove(self, tool_ids: List[str]):
        for tool_id in tool_ids:
            self.tool_store.pop(tool_id)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from pydantic.v1 import BaseModel, Field, Extra

from langchain_pangu.pangukitsappdev.api.common_config import IAMConfig, HttpConfig, OpenAIConfig, IAMConfigWrapper
from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings


class GalleryConfig(SdkBaseSettings):
    gallery_url: Optional[str] = Field(env="sdk.llm.gallery.url")
    iam_config: IAMConfig = Field(default_factory=IAMConfigWrapper(env_prefix="sdk.llm.gallery.iam").get_iam_config)
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.llm.gallery.proxy"))


class LLMParamConfig(BaseModel):
    class Config:
        extra = Extra.forbid

    """
    The maximum number of tokens to generate in the completion.
    """
    max_tokens: Optional[int]

    """
    What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make the output more random, 
    while lower values like 0.2 will make it more focused and deterministic. 
    We generally recommend altering this or top_p but not both.
    """
    temperature: Optional[float]

    """
    An alternative to sampling with temperature, called nucleus sampling, 
    where the model considers the results of the tokens with top_p probability mass. 
    So 0.1 means only the tokens comprising the top 10% probability mass are considered. 
    We generally recommend altering this or temperature but not both.
    """
    top_p: Optional[float]

    """
    How many completions to generate for each prompt. 
    Note: Because this parameter generates many completions, it can quickly consume your token quota. 
    Use carefully and ensure that you have reasonable settings for max_tokens and stop.
    """
    n: Optional[int]

    """
    Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, 
    increasing the model's likelihood to talk about new topics.
    """
    presence_penalty: Optional[float]

    """
    Number between -2.0 and 2.0. 
    Positive values penalize new tokens based on their existing frequency in the text so far, 
    decreasing the model's likelihood to repeat the same line verbatim.
    """
    frequency_penalty: Optional[float]

    """
    Generates best_of completions server-side and returns the "best" (the one with the highest log probability\
     per token).
     Results cannot be streamed. When used with n, best_of controls the number of candidate completions and n specifies\
      how many to return – best_of must be greater than n. 
     Note: Because this parameter generates many completions, it can quickly consume your token quota. 
     Use carefully and ensure that you have reasonable settings for max_tokens and stop.
    """
    best_of: Optional[int]

    """
    If set, partial message deltas will be sent, like in ChatGPT. 
    Tokens will be sent as data-only server-sent events as they become available.
    """
    stream: Optional[bool]

    """
    是否由调用方提供完整prompt，可选参数，默认不设置
    """
    with_prompt: Optional[bool]


DEFAULT_LLM_MODULE_CONFIG_ENV_PREFIX = "sdk.llm.pangu"


class LLMModuleProperty(BaseModel):
    """
    Pangu Agent prompt标志
    Attributes:
        unify_tag_prefix: 输入prompt 起始占位符
        unify_tag_suffix: 输入prompt 结束占位符
        unify_tool_tag_prefix: 工具调用起始占位符
        unify_tool_tag_suffix: 工具调用结束占位符
    """
    unify_tag_prefix: Optional[str]
    unify_tag_suffix: Optional[str]
    unify_tool_tag_prefix: Optional[str]
    unify_tool_tag_suffix: Optional[str]


class LLMModuleConfig(SdkBaseSettings):
    """
    Pangu LLM的基本配置参数
    Attributes:
        llm_name: 模型名称
        url: 模型url
        system_prompt: 系统人设
        enable_append_system_message: 当设置了systemPrompt后，是否尝试自动添加一个SystemMessage，
        在Agent场景下，如果systemPrompt已经拼接在UserMessage了，则会设置为false，不再添加新的systemMessage
        module_version: 盘古模型版本
        cot_desc: cot描述
    """

    def __init__(self, env_prefix=DEFAULT_LLM_MODULE_CONFIG_ENV_PREFIX, **kwargs):
        super().__init__(env_prefix=env_prefix, **kwargs)

    llm_name: str = Field(default="pangu_llm")
    url: Optional[str] = Field(env="url")
    system_prompt: Optional[str]
    enable_append_system_message: bool = True
    module_version: Optional[str] = Field(env="model-version")
    llm_module_property: LLMModuleProperty = Field(default_factory=LLMModuleProperty)
    cot_desc: Optional[str]


class LLMConfig(SdkBaseSettings):
    """LLM参数

    Tips: 这里嵌套的对象，需要使用default_factory，而不是default
    Attributes:
        llm_module_config: llm站点配置，LLMModuleConfig
        iam_config: iam相关认证配置IAMConfig
        llm_param_config: iam认证相关配置LLMParamConfig，默认读取sdk.llm.iam开头的配置
        openai_config: openai认证，OpenAIConfig
        gallery_config: 第三方大模型站点配置，GalleryConfig
        http_config: http相关配置，HttpConfig
    """
    llm_module_config: LLMModuleConfig = Field(default_factory=LLMModuleConfig)
    iam_config: IAMConfig = Field(default_factory=IAMConfigWrapper(env_prefix="sdk.llm.pangu.iam").get_iam_config)
    llm_param_config: LLMParamConfig = Field(default_factory=LLMParamConfig)
    openai_config: OpenAIConfig = Field(default_factory=OpenAIConfig)
    gallery_config: GalleryConfig = Field(default_factory=GalleryConfig)
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.llm.pangu.proxy"))



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging
from abc import ABC, abstractmethod
from typing import List, Any, Optional, Iterable

from langchain.schema import Document as BaseDocument

try:
    from langchain.vectorstores import VectorStore as BaseVectorStore
except ImportError:
    from langchain.schema.vectorstore import VectorStore as BaseVectorStore
from langchain_pangu.pangukitsappdev.api.memory.vector.vector_config import VectorStoreConfig
from langchain_pangu.pangukitsappdev.vectorstores.bulk_data import BulkData

logger = logging.getLogger(__name__)


class Document(BaseDocument):
    """继承langchain.schema.Document
    增加了打分字段
    Attributes:
        score: 文档的相似度打分
        id: 文档id
    """
    score: float = 0.0
    id: str = ""


class VectorStore(BaseVectorStore, ABC):
    @abstractmethod
    def add_docs(self, bulk_data_list: List[BulkData], **kwargs):
        """写入文档统一接口
        Args:
            bulk_data_list: 批量添加的文档
        """

    def add_texts(self, texts: Iterable[str], metadatas: Optional[List[dict]] = None, **kwargs: Any) -> List[str]:
        raise NotImplementedError("Replace add_texts interface with add_docs interface!")

    @abstractmethod
    def clear(self):
        """清空索引文档"""

    @abstractmethod
    def dsl_search(self, dsl: dict) -> dict:
        """通过DSL查询
        Args:
            dsl: DSL
        Returns:
            原始返回结果
        """


class VectorApi(ABC):
    """Interface for vector store"""
    @abstractmethod
    def add_docs(self, bulk_data_list: List[BulkData], **kwargs):
        """写入文档统一接口
        Args:
            bulk_data_list: 批量添加的文档
        """
        pass

    @abstractmethod
    def similarity_search(self, query: str, top_k: int = 5, score_threshold: float = -1.0) -> List[Document]:
        """相似性检索
        在向量库中检索和query相似的文档
        Args:
            query: 查询文本
            top_k: 返回不超过top_k的document，默认5
            score_threshold: 得分阈值，默认-1.0

        Returns:
            一个Document list

        """
        pass

    @abstractmethod
    def search(self, dsl: str) -> dict:
        """通过DSL查询
        Args:
            dsl: 查询语句
        Returns:
            原始返回结果
        """

    @abstractmethod
    def remove(self, ids: List[str]):
        """删除文档
        Args:
            ids: 需要删除的文档ID
        """
        pass

    @abstractmethod
    def clear(self):
        """清空索引文档"""
        pass


class AbstractVectorApi(VectorApi, ABC):
    """VectorApi接口的基类
    封装了一个langchain.vectorstores.VectorStore的实现类，用来适配VectorApi接口。
    子类需要实现create_vector_store，用来构造一个langchain.vectorstores.VectorStore的实现类
    Attributes:
        vector_config: VectorStoreConfig类型，封装了用来构造VectorStore的一些参数
        vector_store: langchain.vectorstores.VectorStore的实现类，通过create_vector_store方法构造
    """

    def __init__(self, vector_config: VectorStoreConfig):
        self.vector_config = vector_config
        self.vector_store: VectorStore = self.create_vector_store(vector_config)

    def add_docs(self, bulk_data_list: List[BulkData], **kwargs):
        self.vector_store.add_docs(bulk_data_list=bulk_data_list, **kwargs)

    def similarity_search(self, query: str, top_k: int = 5, score_threshold: float = -1.0, **kwargs) -> List[Document]:
        return self.vector_store.similarity_search(query=query, k=top_k, score_threshold=score_threshold)

    def search(self, dsl: dict) -> dict:
        return self.vector_store.dsl_search(dsl)

    def remove(self, ids: List[str]):
        self.vector_store.delete(ids=ids)

    def clear(self):
        self.vector_store.clear()

    @abstractmethod
    def create_vector_store(self, vector_config: VectorStoreConfig) -> VectorStore:
        """使用VectorStoreConfig里的配置构造VectorStore实现类

        Args:
            vector_config: 相关配置

        Returns:
            实现类VectorStore接口的对象
        """
        pass



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Any, List

from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import BaseMessage, LLMResult


class StreamCallbackHandler(BaseCallbackHandler):
    def on_chat_model_start(
            self,
            serialized: Dict[str, Any],
            messages: List[List[BaseMessage]],
            **kwargs: Any,
    ) -> Any:
        """
        对话模型开始时运行
        :param serialized: 序列化model
        :param messages: 对话信息
        """

    def on_llm_error(
            self,
            error: BaseException,
            **kwargs: Any,
    ) -> Any:
        """
        对话模型出错时运行
        :param error: 异常信息
        """

    def on_llm_end(
            self,
            response: LLMResult,
            **kwargs: Any,
    ) -> Any:
        """
        对话模型结束时运行
        :param response: 模型结果
        """

    def on_llm_new_token(
            self,
            token: str,
            **kwargs: Any,
    ) -> Any:
        """
        流式每个新token生成时调用
        :param token: 流式生成token
        """



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from langchain_pangu.pangukitsappdev.tool.python_sandbox import PythonSandBox
from langchain_pangu.pangukitsappdev.tool.tool import Tool


class UnifyPythonInterpreterTool(Tool):

    name = "python_interpreter"
    description = ""
    input_desc = ""
    output_desc = ""
    principle = ""
    python_sandbox: Optional[PythonSandBox]
    pangu_function = '{"name": "python_interpreter", "description:": "python解释器",' \
                     ' "principle:": "问题可用编程解决时使用",' \
                     ' "arguments": "String: python代码", "results": "String: 执行结果"}'

    def _run(self, code: str) -> str:
        if self.python_sandbox is None:
            raise ValueError("Param python_sandbox must be specified")
        return self.python_sandbox.run(code)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging
from typing import List

import requests

from langchain_pangu.pangukitsappdev.api.memory.vector.base import Document
from langchain_pangu.pangukitsappdev.api.retriever.retriever_config import WebSearchConfig
from langchain_pangu.pangukitsappdev.utilities.web_search import WebSearch

logger = logging.getLogger(__name__)


class PetalSearch(WebSearch):

    def __init__(self, config: WebSearchConfig):
        """
        初始化
        :param config: 连接信息配置
        """
        self.config = config

    def query(self, query: str, top_k: int) -> List[Document]:
        resp = self.search(query, top_k)
        if resp is None:
            return []

        docs = [Document(page_content=web_page.get('content'),
                         metadata=web_page) for web_page in resp.get("web_pages")]
        return docs

    def search(self, query: str, top_k: int):
        """
        查询
        :param query: 查询问题
        :param top_k: 查询记录数
        :return: 相关网页信息
        """
        proxies = self.config.http_config.requests_proxies()
        requests_url = f"{self.config.server_info.url}?query={query}&limit={top_k}"
        # 账号密码配置
        headers = {"X-HW-ID": self.config.server_info.user,
                   "X-HW-APPKEY": self.config.server_info.password} \
            if self.config.server_info.password is not None else {}
        rsp = requests.get(requests_url, headers=headers, verify=False, proxies=proxies)
        if 200 == rsp.status_code:
            return rsp.json()
        else:
            logger.error("css petal search url error, http status: %d, error response: %s",
                         rsp.status_code, rsp.content)
            return None



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from pydantic.v1 import Field

from langchain_pangu.pangukitsappdev.api.common_config import HttpConfig, ServerInfo
from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings


class KGConfig(SdkBaseSettings):
    server_info: ServerInfo = Field(default_factory=lambda: ServerInfo(env_prefix="sdk.retriever.kg"))
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.retriever.kg.proxy"))


class WebSearchConfig(SdkBaseSettings):
    server_info: ServerInfo = Field(default_factory=lambda: ServerInfo(env_prefix="sdk.retriever.petalSearch"))
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.retriever.petalSearch.proxy"))



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from __future__ import unicode_literals

import json
import logging
from json import JSONDecodeError

from pydantic.v1 import ValidationError

from langchain_pangu.pangukitsappdev.agent.agent_action import AgentAction
from langchain_pangu.pangukitsappdev.agent.agent_session import AgentSession
from langchain_pangu.pangukitsappdev.api.agent.base import AbstractAgent, AgentSessionHelper
from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates

logger = logging.getLogger(__name__)


class ReactAgent(AbstractAgent):
    PLACEHOLDER_THOUGHT = "Thought:"
    PLACEHOLDER_ACTION = "Action:"
    PLACEHOLDER_OBSERVATION = "Observation:"
    DEFAULT_SYS_PROMPT = "您是一个智能助手，尽可能有帮助和准确地回答用户提出的问题"

    def __init__(self, llm: LLMApi):
        super(ReactAgent, self).__init__(llm)

    def react(self, agent_session: AgentSession):
        # 超过最大迭代次数限制，不再执行
        if self.need_interrupt(agent_session):
            return

        # 构造React prompt
        react_tp = PromptTemplates.get("agent_react")

        messages = agent_session.messages
        sys_prompt = self.get_system_prompt(
            agent_session) if self.get_system_prompt(agent_session) is not None else self.DEFAULT_SYS_PROMPT
        final_prompt = react_tp.format(sys_prompt=sys_prompt,
                                       tool_desc=self.get_tool_desc(),
                                       tool_names=self.get_tool_names(),
                                       messages=self.convert_message_to_dict(messages))
        # 获取action
        if self.llm.get_llm_config().llm_param_config.stream:
            tokens = self.llm.ask(final_prompt)
            answer = ""
            for token in tokens:
                answer += token
        else:
            answer = self.llm.ask(final_prompt).answer
        action = self.get_action(answer)
        action.req = final_prompt
        action.resp = answer
        agent_session.current_action = action

        # 如果没有结束，执行action
        if action.action != AgentSessionHelper.FINAL_ACTION:
            # 查询tool
            tool = self.get_tool(action.action)
            if tool is None:
                logger.debug("can not find tool for %s in %s", action.action, str(self.tool_map.keys()))
                raise ValueError("agent did not return a valid action")

            # 执行工具
            try:
                self.tool_execute(tool, action.action_input, agent_session)
            except ValidationError:
                logger.debug("agent did not return a valid tool input, input=%s, but tool need %s",
                             action.action_input, tool.input_type)
                raise ValueError("agent did not return a valid tool input")

            logger.info("actions = %s", "\n".join([action.json(ensure_ascii=False)
                                                   for action in agent_session.current_message.actions]))
        else:
            self.notice_session_end(agent_session, action)

    def get_action(self, answer: str) -> AgentAction:
        # 获取第一个action
        action_str = self.sub_str_before(self.sub_str_after(answer, self.PLACEHOLDER_ACTION),
                                         self.PLACEHOLDER_OBSERVATION)
        if not action_str:
            action_str = self.sub_str_before(answer, self.PLACEHOLDER_OBSERVATION)
        agent_action = self.get_agent_action(action_str)
        agent_action.thought = self.remove_start(self.sub_str_before(answer, self.PLACEHOLDER_ACTION),
                                                 self.PLACEHOLDER_THOUGHT).replace("\n", "")
        return agent_action

    @staticmethod
    def get_agent_action(action_str: str) -> AgentAction:
        try:
            action_json = action_str
            agent_action = AgentAction(**json.loads(action_json.replace("actionInput", "action_input")))
        except (JSONDecodeError, TypeError):
            logger.debug("try to load json failed, json:%s", action_json)
            # 尝试通过工程修复一些LLM返回的错误JSON，修正中文逗号
            fix_str = action_str.replace("，", ",")
            try:
                action_json = fix_str
                agent_action = AgentAction(**json.loads(action_json.replace("actionInput", "action_input")))
            except (JSONDecodeError, TypeError):
                logger.debug("After replace the comma, try to load json failed, json:%s", action_json)
                # 查询一个完整的{}，忽略字符串中的{}(简化)
                first_pos = fix_str.find("{")
                if first_pos < 0 or len(fix_str) < 2:
                    raise ValueError("try to fix json failed, agent did not return a valid action")
                match_count = 1
                last_pos = first_pos + 1
                while last_pos < len(fix_str):
                    if fix_str[last_pos] == "{":
                        match_count += 1
                    if fix_str[last_pos] == "}":
                        match_count -= 1
                    if match_count == 0:
                        break
                    last_pos += 1
                try:
                    action_json = fix_str[first_pos: last_pos + 1]
                    agent_action = AgentAction(**json.loads(action_json.replace("actionInput", "action_input")))
                except (JSONDecodeError, TypeError):
                    logger.debug("After fixed json, try to load json failed, json:%s", action_json)
                    raise ValueError("After fixed json, try to load json failed, json:" + action_json)
        agent_action.action_json = action_json
        return agent_action

    def get_tool_desc(self) -> str:
        return PromptTemplates.get("agent_tool_desc").format(tools=[{"toolId": tool.get_tool_id(),
                                                                     "toolDesc": tool.get_tool_desc(),
                                                                     "toolPrinciple": tool.principle,
                                                                     "inputSchema": tool.get_input_schema(),
                                                                     "outputSchema": tool.get_output_schema()}
                                                                    for tool in self.tool_map.values()])

    def get_tool_names(self) -> str:
        return ", ".join(self.tool_map.keys()) + ", " + AgentSessionHelper.FINAL_ACTION



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import importlib
import os
from typing import Dict, Any, Optional, Mapping, Tuple, Callable

from pydantic.v1 import BaseSettings, Extra
from pydantic.v1.env_settings import SettingsError, DotenvType, SettingsSourceCallable, EnvSettingsSource
from pydantic.v1.utils import deep_update

SDK_CONFIG_PATH_ENV = "SDK_CONFIG_PATH"
ENCRYPTED_PREFIX = "{Crypto."
SDK_CRYPTO_PATH_CONFIG = "sdk.crypto.implementation.path"

ConfigDecryptCallable = Callable[[str, str], str]


class SdkBaseSettings(BaseSettings):
    """
    Sdk的配置顶层类，覆盖了Pydantic的一些默认行为。通过在初始化的时候传递env_prefix参数来指定配置加载的key前缀，
    传递env_prefix_ident_fields参数，用于指定env_prefix对指定field生效，默认None对全部field生效
    """

    class Config:
        extra = Extra.ignore
        env_file = os.environ.get(SDK_CONFIG_PATH_ENV)

        @classmethod
        def customise_sources(
                cls,
                init_settings: SettingsSourceCallable,
                env_settings: SettingsSourceCallable,
                file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            env_file = env_settings.env_file if env_settings.env_file else os.environ.get(SDK_CONFIG_PATH_ENV)
            prop_settings = SdkEnvSettingsSource(env_file, env_settings.env_file_encoding,
                                                 env_settings.env_nested_delimiter, env_settings.env_prefix_len,
                                                 init_settings.init_kwargs.get("env_prefix"),
                                                 init_settings.init_kwargs.get("env_name_joint", "."),
                                                 init_settings.init_kwargs.get("env_prefix_ident_fields", None))
            return init_settings, prop_settings, file_secret_settings


class SdkEnvSettingsSource(EnvSettingsSource):
    """
    主干逻辑是__call__方法，结构上基本上和EnvSettingsSource保持一致。区别点：\n
    1. dotenv_vars配置文件的配置优先级高于环境变量.\n
    2. 增加env_prefix成员变量，用来实现匹配不同的前缀做配置加载\n
    3. 增加env_name_joint成员变量，默认使用"."连接env_prefix和env_name\n
    4. 增加config_decryptor类变量用来做配置的的解密，使用者可以自行配置。
    5. 增加env_prefix_ident_fields对指定元素生效，默认None
    """

    config_decryptor: ConfigDecryptCallable = None

    @classmethod
    def load_config_decryptor(cls, import_path: str) -> ConfigDecryptCallable:
        """从import_path加载config_decryptor
        加载自定义的config_decryptor。比如自定义的config_decryptor使用如下：\n
        from your.module import decrypt_func\n
        plain_text = decrypt_func(key_id, cipher)\n

        则等同使用如下方法加载：\n
        decrypt_func = load_config_decryptor("your.module.decrypt_func")\n
        plain_text = decrypt_func(key_id, cipher)\n

        该方法用来初始化类变量：SdkEnvSettingsSource.config_decryptor。只会初始化一次
        Args:
            import_path: 自定义config_decryptor的路径，基本格式，{path.to.module}.{function_name}

        Returns:
            一个可调用的方法对象

        """

        if cls.config_decryptor:
            return

        module_path, func_name = import_path.rsplit(".", 1)
        m = importlib.import_module(module_path)
        f = getattr(m, func_name)
        cls.config_decryptor = f

    __slots__ = (
        'env_file', 'env_file_encoding', 'env_nested_delimiter', 'env_prefix_len', 'env_prefix', 'env_name_joint',
        'env_prefix_ident_fields')

    def __init__(
            self,
            env_file: Optional[DotenvType],
            env_file_encoding: Optional[str],
            env_nested_delimiter: Optional[str] = None,
            env_prefix_len: int = 0,
            env_prefix: str = None,
            env_name_joint: str = ".",
            env_prefix_ident_fields: Optional[list] = None
    ):
        self.env_file: Optional[DotenvType] = env_file
        self.env_file_encoding: Optional[str] = env_file_encoding
        self.env_nested_delimiter: Optional[str] = env_nested_delimiter
        self.env_prefix_len: int = env_prefix_len
        # 在初始化的时候传入env_prefix，从环境变量中取值的时候考虑使用，默认是None
        self.env_prefix = env_prefix
        self.env_name_joint = env_name_joint
        self.env_prefix_ident_fields = env_prefix_ident_fields

    def __call__(self, settings: BaseSettings) -> Dict[str, Any]:  # noqa C901
        """
        Build environment variables suitable for passing to the Model.
        """
        d: Dict[str, Any] = {}

        if settings.__config__.case_sensitive:
            env_vars: Mapping[str, Optional[str]] = os.environ
        else:
            env_vars = {k.lower(): v for k, v in os.environ.items()}

        dotenv_vars = self._read_env_files(settings.__config__.case_sensitive)
        if dotenv_vars:
            """这里修改了，优先使用dotenv_vars里的配置，要求文件配置优先级高于环境变量配置，支持配置文件留空
            """
            dotenv_vars = {k: (dotenv_vars.get(k) if dotenv_vars.get(k) else None) for k in dotenv_vars.keys()}
            env_vars = {**env_vars, **dotenv_vars}

        # 加载load_config_decryptor
        sdk_crypto_path = env_vars.get(SDK_CRYPTO_PATH_CONFIG)
        if sdk_crypto_path:
            self.load_config_decryptor(sdk_crypto_path)

        for field in settings.__fields__.values():
            env_val: Optional[str] = None
            for env_name in field.field_info.extra['env_names']:
                need_prefix = self.env_prefix_ident_fields is None or env_name in self.env_prefix_ident_fields
                if self.env_prefix and need_prefix:
                    env_name = self.env_prefix.lower() + self.env_name_joint + env_name
                env_val = self.decrypt_if_need(env_vars.get(env_name))
                if env_val is not None:
                    break

            is_complex, allow_parse_failure = self.field_is_complex(field)
            if is_complex:
                if env_val is None:
                    # field is complex but no value found so far, try explode_env_vars
                    env_val_built = self.explode_env_vars(field, env_vars)
                    if env_val_built:
                        d[field.alias] = env_val_built
                else:
                    # field is complex and there's a value, decode that as JSON, then add explode_env_vars
                    try:
                        env_val = settings.__config__.parse_env_var(field.name, env_val)
                    except ValueError as e:
                        if not allow_parse_failure:
                            raise SettingsError(f'error parsing env var "{env_name}"') from e

                    if isinstance(env_val, dict):
                        d[field.alias] = deep_update(env_val, self.explode_env_vars(field, env_vars))
                    else:
                        d[field.alias] = env_val
            elif env_val is not None:
                # simplest case, field is not complex, we only need to add the value if it was found
                d[field.alias] = env_val

        return d

    @classmethod
    def decrypt_if_need(cls, config_value: str) -> str:
        if not cls.config_decryptor:
            return config_value

        if config_value and config_value.startswith(ENCRYPTED_PREFIX):
            key_id_end_idx = config_value.index("}")
            key_id = config_value[len(ENCRYPTED_PREFIX):key_id_end_idx]
            encrypted_value = config_value[key_id_end_idx + 1:]
            return cls.config_decryptor(key_id, encrypted_value)
        else:
            return config_value



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from pydantic.v1 import Field

from langchain_pangu.pangukitsappdev.api.common_config import ServerInfo
from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings
from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import ServerInfoRedis


class MessageHistoryConfig(SdkBaseSettings):
    """
    缓存存储的相关配置
    Attributes:
        store_name: 存储名称
        server_info: ServerInfo，存储服务器连接相关信息，默认使用redis的连接
        key_prefix: redis key前缀
        table_name: sql 表名称
        ttl: 缓存的老化时间，单位 秒
        session_tag: 用户指定history会话标志
    """

    store_name: Optional[str]
    server_info: ServerInfo = Field(default_factory=lambda: ServerInfoRedis(env_prefix="sdk.memory.dcs"))
    key_prefix: str = Field(default="message_store:")
    table_name: str = Field("tbl_chat_message_history")
    ttl: int = Field(default=24 * 3600)
    session_tag: str = Field(default="")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging
from typing import List

import requests

from langchain_pangu.pangukitsappdev.api.memory.vector.base import Document
from langchain_pangu.pangukitsappdev.api.retriever.retriever_config import KGConfig
from langchain_pangu.pangukitsappdev.utilities.knowledge_graph import KnowledgeGraph
from langchain_pangu.pangukitsappdev.utils.security_util import SecurityUtil
from langchain_pangu.pangukitsappdev.utils.time_date import now_millis

logger = logging.getLogger(__name__)


class KG(KnowledgeGraph):
    def __init__(self, config: KGConfig):
        """
        初始化
        :param config: 连接信息配置
        """
        self.config = config

    def query(self, query: str) -> List[Document]:
        resp = self.search(query)
        if resp is None:
            return []

        docs = [Document(page_content=resp.get('avoice'), metadata=resp)]
        return docs

    def search(self, query: str):
        """
        查询
        :param query: 查询问题
        :return: 对应结果
        """
        requests_body = {
            "queryString": query,
            "locale": "zh-cn"
        }
        logger.debug("css kg request: %s", str(requests_body))
        proxies = self.config.http_config.requests_proxies()
        headers = self.set_auth()
        headers.update({"Content-Type": "application/json"})
        rsp = requests.post(url=self.config.server_info.url,
                            json=requests_body,
                            headers=headers,
                            verify=False,
                            proxies=proxies)

        logger.debug("css kg request: %s", str(rsp))
        if 200 == rsp.status_code:
            return rsp.json()
        else:
            logger.error("css petal search url error, http status: %d, error response: %s",
                         rsp.status_code, rsp.content)
            return None

    def set_auth(self) -> dict:
        if self.config.server_info.password is None:
            return {}
        nonce = SecurityUtil.get_security_random()
        timestamp = now_millis()
        data = str(timestamp) + nonce
        auth = SecurityUtil.hmac_sha256_base64(data, self.config.server_info.password)
        return {
            "nonce": nonce,
            "timestamp": str(timestamp),
            "AuthCode": auth
        }



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import List

import requests
from langchain.schema.embeddings import Embeddings
from pydantic.v1 import BaseModel

from langchain_pangu.pangukitsappdev.api.embeddings.base import AbstractEmbeddingApi
from langchain_pangu.pangukitsappdev.api.embeddings.embedding_config import EmbeddingConfig


class ToolEmbeddings(Embeddings, BaseModel):
    embedding_config: EmbeddingConfig

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        return self.create_embedding(texts)

    def embed_query(self, text: str) -> List[float]:
        """对query词做Embedding

        检索时对查询词做Embedding

        :param text: 查询词
        :return: :class:`List[float] <List[float]>` 向量数据
        :rtype: List[float]
        """
        return self.embed_documents([text])[0]

    def create_embedding(self, texts: List[str]) -> List[List[float]]:
        proxies = self.embedding_config.http_config.requests_proxies()
        request_body = {
            "content": texts,
            "function": "embedding"
        }

        headers = {"Content-Type": "application/json"}

        rsp = requests.post(url=self.embedding_config.css_url,
                            headers=headers,
                            json=request_body,
                            proxies=proxies,
                            verify=False)
        result = [[]]
        if 200 == rsp.status_code:
            result = rsp.json().get('content')
        else:
            rsp.raise_for_status()
        return result


class ToolEmbeddingApi(AbstractEmbeddingApi):

    def create_embeddings(self, embedding_config: EmbeddingConfig) -> Embeddings:
        return ToolEmbeddings(embedding_config=embedding_config)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional, Type

from pydantic.v1 import Field, BaseModel

from langchain_pangu.pangukitsappdev.tool.python_sandbox import PythonSandBox
from langchain_pangu.pangukitsappdev.tool.tool import Tool


class PythonInterpreterTool(Tool):
    class InputParam(BaseModel):
        code: str = Field(description="Python代码")

    name = "python_interpreter"
    description = "通过调用python解释器执行代码来解决计算和文件操作类问题"
    principle = "请在需要进行数学计算、日期查询或者文件操作等场景使用此工具"
    args_schema: Type[BaseModel] = InputParam
    output_desc = "代码执行结果"
    python_sandbox: Optional[PythonSandBox]

    def _run(self, code: str) -> str:
        if self.python_sandbox is None:
            raise ValueError("Param python_sandbox must be specified")
        return self.python_sandbox.run(code)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain_pangu.pangukitsappdev.api.schema import LLMResp
from langchain_pangu.pangukitsappdev.llms.response.gallery_text_resp import GalleryTextResp


class LLMRespGallery(LLMResp):
    """gallery response封装结构体
    Attributes:
        gallery_text_resp: api返回response结构体
    """
    gallery_text_resp: GalleryTextResp



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import List, Optional

from pydantic.v1 import BaseModel


class GalleryTextChoice(BaseModel):
    """第三方模型生成的补全信息的列表
    Attributes:
        index: 补全的索引，从0开始
        text: 补全的文本内容
    """
    index: Optional[int]
    text: Optional[str]


class GalleryUsage(BaseModel):
    """第三方模型资源使用情况
    Attributes:
        completion_tokens: 表示模型生成的答案中包含的tokens的数量
        prompt_tokens: 表示生成结果时使用的提示文本的tokens的数量
        total_tokens: 对话过程中使用的tokens总数
    """
    completion_tokens: Optional[int]
    prompt_tokens: Optional[int]
    total_tokens: Optional[int]


class GalleryTextResp(BaseModel):
    """第三方模型返回的response
    Attributes:
        id: 用来标识每个响应的唯一字符串
        created: 响应生成的时间
        choices: 生成的补全信息的列表
        usage: 模型资源使用情况
    """
    id: Optional[str]
    created: Optional[str]
    choices: Optional[List[GalleryTextChoice]]
    usage: Optional[GalleryUsage]



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from abc import abstractmethod, ABC
from typing import List

from langchain_pangu.pangukitsappdev.api.tool.base import AbstractTool
from langchain_pangu.pangukitsappdev.retriever.retrieved_tool import RetrievedTool


class ToolProvider(ABC):
    """工具实例化提供者，通过toolId列表给出实例化工具
    """
    @abstractmethod
    def provide(self, retrieved_tools: List[RetrievedTool], query: str) -> List[AbstractTool]:
        """通过id查找工具
        Args:
            retrieved_tools: 通过ToolRetriever召回的工具
            query: 查询语句
        Returns: 实例化后的工具
        """



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import json
import logging
from typing import Optional, List

from langchain_pangu.pangukitsappdev.retriever.web_retirever import WebRetriever
from langchain_pangu.pangukitsappdev.tool.tool import Tool
from langchain_pangu.pangukitsappdev.utilities.web_search import WebSearch

logger = logging.getLogger(__name__)


class UnifyWebSearch(Tool):

    name = "web_search"
    description = ""
    input_desc = ""
    output_desc = ""
    principle = ""
    web_search: Optional[WebSearch]
    pangu_function = '{"name": "web_search", "description:": "搜索引擎获取互联网知识",' \
                     ' "principle:": "问题可用搜索引擎辅助解决时使用",' \
                     ' "arguments": {"query": ["String:搜索query", "..."]}, "results": "搜索结果片段列表"}'

    def _run(self, query_list: List[str]) -> str:
        if self.web_search is None:
            raise ValueError("Param web_search must be specified")
        web_retriever = WebRetriever(web_search=self.web_search)
        query = " ".join(query_list)
        logger.info("[web_search工具调用问题]：%s", query)
        docs = web_retriever.get_relevant_documents(query=query, top_k=5)

        if len(docs) > 0:
            try:
                return json.dumps({"evidence": [{"category": doc.metadata.get("site_category", ["无"])[0],
                                                 "publish_time": doc.metadata.get("publish_time", "无"),
                                                 "text": doc.page_content,
                                                 "title": doc.metadata.get("title", "无")} for doc in docs],
                                   "recent_user_query": query}, ensure_ascii=False)
            except Exception as e:
                logger.error("出现如下异常%s", e)
                return "搜索信息错误，不要调用工具，直接回答问题。"

        else:
            return "搜索信息错误，不要调用工具，直接回答问题。"



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import datetime
import time

F_yyyyMMddHHmmss = "%Y-%m-%d %H:%M:%S"

def now_millis() -> int:
    """
    返回当前的时间戳毫秒值
    :return: now毫秒值
    """
    return round(time.time() * 1000)


def now_sec() -> int:
    """返回当前时间戳，单位秒"""
    return int(time.time())


def now_yyyyMMddHHmmss() -> str:
    """
    以yyyy-MM-dd HH:mm:ss 格式返回当前的时间
    Returns:
        yyyy-MM-dd HH:mm:ss格式的时间

    """

    return datetime.datetime.now().strftime(F_yyyyMMddHHmmss)


def to_yyyyMMddHHmmss(ts_sec: int) -> str:
    """
    把ts_sec转换成yyyy-MM-dd HH:mm:ss 格式字符串
    Args:
        ts_sec: 秒表示的时间戳

    Returns:
        yyyy-MM-dd HH:mm:ss 格式时间戳

    """
    time_arr = time.localtime(ts_sec)
    return time.strftime(F_yyyyMMddHHmmss, time_arr)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import json
from typing import List

from langchain.schema import BaseChatMessageHistory
from langchain.schema.messages import BaseMessage, messages_from_dict
from sqlalchemy import Column, Integer, Text, create_engine, String
from sqlalchemy.orm import sessionmaker

from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import ServerInfoSql
from langchain_pangu.pangukitsappdev.api.memory.message_history_config import MessageHistoryConfig
from langchain_pangu.pangukitsappdev.utils.time_date import now_yyyyMMddHHmmss, now_sec, to_yyyyMMddHHmmss


def message_to_dict(message: BaseMessage) -> dict:
    return {"type": message.type, "data": message.dict()}


def create_message_model_with_create_date(table_name, dynamic_base):
    """
    基于表名创建 message model 对比 langchain SQLChatMessageHistory新增 create_date
    Args:
        table_name: The name of the table to use.
        dynamic_base: The base class to use for the model.
    Returns:
        The model class.
    """

    # Model declared inside a function to have a dynamic table name
    class Message(dynamic_base):
        __tablename__ = table_name
        id = Column(Integer, primary_key=True)
        session_tag = Column(Text)
        message = Column(Text)
        create_date = Column(String(45), nullable=True, default=None, index=True)

    return Message


class SQLMessageHistory(BaseChatMessageHistory):
    """SQLMessageHistory重写"""

    def __init__(self, msg_history_config: MessageHistoryConfig = None):
        if not msg_history_config:
            msg_history_config = MessageHistoryConfig(server_info=ServerInfoSql(env_prefix="sdk.memory.rds"))
        if not isinstance(msg_history_config.server_info, ServerInfoSql):
            raise ValueError("Type of server_info should be ServerInfoSql")
        self.table_name = msg_history_config.table_name
        self.connection_string = msg_history_config.server_info.get_urls()[0]
        self.engine = create_engine(self.connection_string, echo=False)
        self._create_table_if_not_exists()
        self.session_tag = msg_history_config.session_tag
        self.Session = sessionmaker(self.engine)
        self.ttl = msg_history_config.ttl

    def _create_table_if_not_exists(self) -> None:
        try:
            from sqlalchemy.orm import declarative_base
        except ImportError:
            from sqlalchemy.ext.declarative import declarative_base
        dynamic_base = declarative_base()
        self.Message = create_message_model_with_create_date(self.table_name, dynamic_base)
        # Create all does the check for us in case the table exists.
        dynamic_base.metadata.create_all(self.engine)

    @property
    def messages(self) -> List[BaseMessage]:
        """返回所有 session_tag下 message"""
        with self.Session() as session:
            result = session.query(self.Message).where(
                self.Message.session_tag == self.session_tag
            )
            items = [json.loads(record.message) for record in result]
            messages = messages_from_dict(items)
            return messages

    def add_message(self, message: BaseMessage) -> None:
        """新增 message 前移除过期 message"""
        self.evict()
        with self.Session() as session:
            json_str = json.dumps(message_to_dict(message))
            session.add(self.Message(
                session_tag=self.session_tag,
                message=json_str,
                create_date=now_yyyyMMddHHmmss()
            ))
            session.commit()

    def clear(self) -> None:
        """删除 session_tag 所有 message"""
        with self.Session() as session:
            session.query(self.Message).filter(
                self.Message.session_tag == self.session_tag
            ).delete()
            session.commit()

    def evict(self) -> None:
        """根据 ttl 删除过期 message"""
        with self.Session() as session:
            session.query(self.Message).filter(
                self.Message.session_tag == self.session_tag,
                self.Message.create_date < to_yyyyMMddHHmmss(now_sec() - self.ttl)
            ).delete()
            session.commit()



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from pydantic.v1 import BaseModel


class OpenAIUsage(BaseModel):
    """OpenAI模型资源使用情况
    Attributes:
        completion_tokens: 表示模型生成的答案中包含的tokens的数量
        prompt_tokens: 表示生成结果时使用的提示文本的tokens的数量
        total_tokens: 对话过程中使用的tokens总数
    """
    completion_tokens: Optional[int]
    prompt_tokens: Optional[int]
    total_tokens: Optional[int]


class OpenAITextResp(BaseModel):
    """OpenAI模型返回的response
    Attributes:
        model: 模型名
        usage: 模型资源使用情况
    """
    model: Optional[str]
    usage: Optional[OpenAIUsage]



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.

from abc import ABC, abstractmethod
from typing import List, Iterator

from langchain.document_loaders.base import BaseLoader
from langchain.schema import Document

from langchain_pangu.pangukitsappdev.api.doc_split.split_config import SplitConfig


class LoaderApi(BaseLoader, ABC):
    @abstractmethod
    def load(self) -> List[Document]:
        """
        对文档进行加载并解析分割
        :return: 分割的document列表
        """
        pass


class AbstractLoaderApi(LoaderApi, ABC):

    def __init__(self, split_config: SplitConfig):
        self.split_config = split_config

    @abstractmethod
    def load(self) -> List[Document]:
        """
        实现类方法
        :return: 解析分割后的document列表
        """
        pass


    def lazy_load(self) -> Iterator[Document]:
        # 默认未实现，直接抛异常
        raise NotImplementedError("Unimplement method!")


#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import json
from typing import Dict

import requests

from langchain_pangu.pangukitsappdev.api.common_config import IAMConfig
from langchain_pangu.pangukitsappdev.api.common_config import IAM_TOKEN_HEADER
from langchain_pangu.pangukitsappdev.utils.time_date import now_millis


class IAMTokenProvider:
    """盘古鉴权的接口
        token = PanGuAuth(url={iam auth url}, username={your username}, password={your password},
        domain_name={your domain_name},project_name={your project_name} )
    """

    def __init__(self, iam_config: IAMConfig) -> None:

        """构造器

        :param iam_config: 相关配置
        """
        self.iam_disabled = iam_config.iam_disabled
        self.iam_ak = iam_config.iam_ak
        self.iam_sk = iam_config.iam_sk
        self.iam_url = iam_config.iam_url
        self.project_name = iam_config.project_name
        self.iam_domain = iam_config.iam_domain
        self.iam_pwd = iam_config.iam_pwd
        self.iam_user = iam_config.iam_user
        self.expire_duration_millis = iam_config.expire_duration_millis
        self.proxies = iam_config.http_config.requests_proxies()
        self.x_auth_token = iam_config.x_auth_token

        # 当前的token，可能过期了
        self.__hold_token: str = ""

        # 在什么时间超时，一个时间戳
        self.__expired_at: int = -1
        # 什么时间创建的
        self.__created_at: int = -1

    @staticmethod
    def get_token(url: str, username: str, password: str, domain_name: str, project_name: str, proxies: dict) -> str:
        data = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': username,
                            'password': password,
                            'domain': {
                                'name': domain_name
                            }
                        }
                    }
                },
                'scope': {
                    'project': {
                        'name': project_name
                    }
                }
            }
        }
        headers = {
            'Content-Type': 'application/json'
        }

        response = requests.post(url, data=json.dumps(data), headers=headers, verify=False, timeout=15, proxies=proxies)
        if response.status_code == 201:
            # 请求成功，处理响应数据
            json_response = response.headers
            token = json_response[IAM_TOKEN_HEADER]
            return token

        response.raise_for_status()

    @staticmethod
    def get_token_by_ak_sk(url: str, access_key: str, secret_key: str, project_name: str, proxies: dict) -> str:
        data = {
            'auth': {
                'identity': {
                    'hw_ak_sk': {
                        'access': {
                            'key': access_key
                        },
                        "secret": {
                            'key': secret_key
                        }
                    },
                    'methods': ['hw_ak_sk']
                },
                'scope': {
                    'project': {
                        'name': project_name
                    }
                }
            }
        }
        headers = {
            'Content-Type': 'application/json'
        }

        response = requests.post(url, data=json.dumps(data), headers=headers, verify=False, timeout=15, proxies=proxies)
        if response.status_code == 201:
            # 请求成功，处理响应数据
            json_response = response.headers
            token = json_response[IAM_TOKEN_HEADER]
            return token

        response.raise_for_status()
        return ""

    def get_valid_token(self):
        """获取有效token
        使用这个方法获取可用token，并在到期后自动刷新
        :return: iam认证token
        """
        if self.iam_disabled:
            return ""

        if self.x_auth_token:
            return self.x_auth_token

        if self.__expired_at and now_millis() > self.__expired_at:
            if self.iam_sk is not None and self.iam_ak is not None:
                fresh_token = IAMTokenProvider.get_token_by_ak_sk(url=self.iam_url,
                                                                  access_key=self.iam_ak,
                                                                  secret_key=self.iam_sk,
                                                                  project_name=self.project_name,
                                                                  proxies=self.proxies)
            else:
                fresh_token = IAMTokenProvider.get_token(url=self.iam_url,
                                                         username=self.iam_user,
                                                         password=self.iam_pwd,
                                                         domain_name=self.iam_domain,
                                                         project_name=self.project_name,
                                                         proxies=self.proxies)
            self._update_token(fresh_token)

        return self.__hold_token

    def _update_token(self, fresh_token):
        self.__hold_token = fresh_token
        self.__created_at = now_millis()
        self.__expired_at = self.__created_at + self.expire_duration_millis


class IAMTokenProviderFactory:
    # 实现享元模式，同样的配置不用每次都创建一次。key是iamconfig的序列化数据
    _instance_cache: Dict[str, IAMTokenProvider] = {}

    @classmethod
    def create(cls, iam_config: IAMConfig) -> IAMTokenProvider:
        config_json_str = iam_config.json()
        cached = cls._instance_cache.get(config_json_str)
        if cached:
            return cached
        else:
            new_one = IAMTokenProvider(iam_config)
            # 更新缓存
            cls._instance_cache[config_json_str] = new_one

            return new_one



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain.chains.llm import LLMChain
from langchain.schema.prompt_template import BasePromptTemplate

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.api.skill.base import ChainWrappedSkill, SimpleSkill


class Skills:
    """
    静态工厂类，用来创建Skill
    """
    @classmethod
    def of_chain(cls, chain: LLMChain) -> ChainWrappedSkill:
        return ChainWrappedSkill(chain)

    @classmethod
    def of(cls, prompt_template: BasePromptTemplate, llm_api: LLMApi):
        return SimpleSkill(prompt_template, llm_api)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.api.skill.base import SimpleSkill
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates


class DocAskSkill(SimpleSkill):
    def __init__(self, llm_api: LLMApi):
        super().__init__(PromptTemplates.get("documents_stuff"), llm_api)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional, List

from pydantic.v1 import BaseModel

from langchain_pangu.pangukitsappdev.agent.agent_action import AgentAction
from langchain_pangu.pangukitsappdev.api.llms.base import ConversationMessage


class AgentSession(BaseModel):
    """
    Agent运行Session，包含历史Action，当前Action，状态
    Attributes:
        messages: 本次session的用户的输入
        session_id: UUID，在一个session内唯一
        current_action: 当前Action
        agent_session_status: Agent状态
        is_by_step: 是否是逐步执行
        current_message: 当前的AssistantMessage
    """
    messages: List[ConversationMessage]
    session_id: str
    current_action: Optional[AgentAction]
    agent_session_status: str
    is_by_step: Optional[bool]
    current_message: Optional[ConversationMessage]



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from pydantic.v1 import Field

from langchain_pangu.pangukitsappdev.api.common_config import IAMConfig, OpenAIConfig, HttpConfig, IAMConfigWrapper
from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings

DEFAULT_EMBEDDING_CONFIG_ENV_PREFIX = "sdk.embedding.css"


class EmbeddingConfig(SdkBaseSettings):

    def __init__(self, env_prefix=DEFAULT_EMBEDDING_CONFIG_ENV_PREFIX, **kwargs):
        super().__init__(env_prefix=env_prefix, **kwargs)

    embedding_name: Optional[str] = Field(default="default_embedding_name")
    css_url: Optional[str] = Field(env="url")
    iam_config: IAMConfig = Field(default_factory=IAMConfigWrapper(env_prefix="sdk.embedding.css.iam").get_iam_config)
    openai_config: OpenAIConfig = Field(default_factory=OpenAIConfig)
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.embedding.css.proxy"))



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain.chat_models.base import BaseChatModel
from langchain.llms.base import BaseLLM
from langchain.llms.openai import OpenAI
from langchain.schema import LLMResult
from langchain_openai import ChatOpenAI

from langchain_pangu.pangukitsappdev.api.llms.base import AbstractLLMApi
from langchain_pangu.pangukitsappdev.api.llms.llm_config import LLMConfig
from langchain_pangu.pangukitsappdev.llms.response.llm_response_openai import LLMRespOpenAI
from langchain_pangu.pangukitsappdev.llms.response.openai_text_resp import OpenAIUsage, OpenAITextResp


class OpenAILLMApi(AbstractLLMApi):
    def do_create_llm(self, llm_config: LLMConfig) -> BaseLLM:
        config_params = self._parse_llm_config(llm_config)
        return OpenAI(**config_params)

    def do_create_chat_llm(self, llm_config: LLMConfig) -> BaseChatModel:
        config_params = self._parse_llm_config(llm_config)
        return ChatOpenAI(**config_params)

    def parse_llm_response(self, llm_result: LLMResult) -> LLMRespOpenAI:
        answer = llm_result.generations[0][0].text
        llm_output = llm_result.llm_output

        if llm_output.get("token_usage"):
            usage = OpenAIUsage(completion_tokens=llm_output.get("token_usage").get("completion_tokens"),
                                prompt_tokens=llm_output.get("token_usage").get("prompt_tokens"),
                                total_tokens=llm_output.get("token_usage").get("total_tokens"))
        else:
            usage = None
        llm_resp = LLMRespOpenAI(answer=answer,
                                 is_from_cache=False,
                                 openai_text_resp=OpenAITextResp(model=llm_output.get("model_name"),
                                                                 usage=usage))
        return llm_resp

    def _parse_llm_config(self, llm_config: LLMConfig) -> dict:
        config_params = {}

        if llm_config.openai_config.openai_base_url:
            config_params["openai_api_base"] = llm_config.openai_config.openai_base_url

        if llm_config.openai_config.openai_key:
            config_params["openai_api_key"] = llm_config.openai_config.openai_key

        llm_param_config_dict: dict = llm_config.llm_param_config.dict(exclude_none=True, exclude={"stream"})
        if llm_config.llm_param_config.stream:
            llm_param_config_dict["streaming"] = llm_config.llm_param_config.stream
        config_params = {**config_params, **llm_param_config_dict}

        # 配置代理
        if llm_config.openai_config.http_config.proxy_enabled:
            config_params["openai_proxy"] = llm_config.openai_config.http_config.get_proxy_url()
        return config_params



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Any

from langchain.memory.buffer import ConversationBufferMemory
from langchain.prompts import PromptTemplate
from langchain.schema.memory import BaseMemory

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.api.skill.base import Skill
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates


class ConversationSkill(Skill):

    class PromptParam:
        MEMORY_HISTORY = "history"
        INPUT = "input"
        OUTPUT = "output"

    def execute_dict(self, inputs: Dict[str, Any]) -> str:
        if self.PromptParam.MEMORY_HISTORY in inputs:
            del inputs[self.PromptParam.MEMORY_HISTORY]
        inputs.update(self.memory.load_memory_variables({}))
        answer = self.skill_llm_ask(self.prompt.format(**inputs), self.llm)
        self.memory.save_context({self.PromptParam.INPUT: inputs.get(self.PromptParam.INPUT)},
                                 {self.PromptParam.OUTPUT: answer})
        return answer

    def __init__(self, llm: LLMApi):
        self.llm = llm
        self.memory = ConversationBufferMemory()
        self.prompt = PromptTemplates.get("conversation_default")

    def set_prompt(self, prompt_template: str):
        self.prompt = PromptTemplate.from_template(prompt_template, template_format="jinja2")

    def set_memory(self, memory: BaseMemory):
        self.memory = memory

    def execute(self, input_str: str) -> str:
        inputs = {self.PromptParam.INPUT: input_str}
        return self.execute_dict(inputs)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from pydantic.v1 import Field

from langchain_pangu.pangukitsappdev.api.common_config import ServerInfo
from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings
from langchain_pangu.pangukitsappdev.api.embeddings.base import EmbeddingApi


class ServerInfoRedis(ServerInfo):

    def has_auth(self) -> bool:
        return True if self.password else False

    def get_urls(self) -> [str]:
        if not self.url:
            return []

        from urllib.parse import quote
        result = []
        for u in self.url.split(","):
            if u.find('://') != -1:
                result.append(f"redis://:{quote(self.password)}@{u.split('://')[1]}/0"
                              if self.has_auth() else f"redis://{u.split('://')[1]}/0")
            else:
                result.append(f"redis://:{quote(self.password)}@{u}/0" if self.has_auth() else f"redis://{u}/0")

        return result


class ServerInfoSql(ServerInfo):
    """sql数据存储相关的配置
    Attributes:
        url: 拼接好的url； 如jdbc场景
        pool_size: 连接池大小
    """

    pool_size: Optional[str] = Field(env="poolSize", default=5)

    def get_urls(self) -> [str]:
        from sqlalchemy.engine.url import make_url
        url_instance = make_url(self.url)

        url_instance = url_instance.set(username=self.user, password=self.password)
        return [url_instance.render_as_string(False)]


class CacheStoreConfig(SdkBaseSettings):
    """
    缓存存储的相关配置
    Attributes:
        store_name: 存储名称
        server_info: ServerInfo，存储服务器连接相关信息，默认使用redis的连接
        embedding: embedding接口，语义缓存时需要使用
        vector_store_name: 向量语义缓存时，设置库名，默认使用css存储
        distance_strategy: 检索模式，默认使用内积inner_product
        score_threshold: 语义缓存， 相似度评分阈值；评分低于阈值表示不相似
        session_tag: 用户指定cache会话标志
        expire_after_access: 访问后到期时间，单位为秒， 默认不设置过期
        expire_after_write: 写入后到期时间，单位为秒， 默认不设置过期
        maximum_size: 基于个数大小
    """

    store_name: Optional[str]
    server_info: ServerInfo = Field(default_factory=lambda: ServerInfoRedis(env_prefix="sdk.memory.dcs"))
    embedding: Optional[EmbeddingApi]
    vector_store_name: str = Field(default="css")
    distance_strategy: str = Field(default="inner_product")
    score_threshold: float = Field(default=0.2)
    session_tag: str = Field(default="")
    expire_after_access: int = Field(default=-1)
    expire_after_write: int = Field(default=-1)
    maximum_size: int = Field(default=-1)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import json
import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Iterator, Type, List, Union, Literal, Optional, Any, Dict

from langchain.callbacks.base import BaseCallbackHandler
from langchain.chat_models.base import BaseChatModel
from langchain.schema import LLMResult, BaseMessage
from pydantic.v1 import BaseModel

from langchain_pangu.pangukitsappdev.agent.agent_action import AgentAction
from langchain_pangu.pangukitsappdev.api.llms.llm_config import LLMParamConfig, LLMConfig
from langchain_pangu.pangukitsappdev.api.memory.cache.base import CacheApi
from langchain_pangu.pangukitsappdev.api.schema import LLMResp
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates

logger = logging.getLogger(__name__)


class Role(Enum):
    SYSTEM = {"text": "system", "desc": "系统"}
    USER = {"text": "user", "desc": "用户"}
    ASSISTANT = {"text": "assistant", "desc": "助手"}
    OBSERVATION = {"text": "observation", "desc": "观察"}

    @property
    def desc(self):
        return self.value.get("desc")

    @property
    def text(self):
        return self.value.get("text")


class ConversationMessage(BaseMessage):
    """多轮对话信息
    拓展了tool和actions，可用于Agent使用

    Attributes:
        role: 对话角色
        content: 对话内容
        tools: 工具集
        actions: 当role为assistant时，采取的actions，如果没有采取任何action，则为空数组
    """
    role: Role
    """The speaker / role of the Message."""

    type: Literal["chat"] = "chat"

    tools: Optional[Any]

    actions: List[AgentAction] = []


def convert_message_to_req(messages: List[BaseMessage]) -> List[Dict]:
    req = []
    for message in messages:
        if not isinstance(message, ConversationMessage):
            raise ValueError("Pangu LLM input message must be inherited from ConversationMessage!")
        req.append({"role": message.role.text, "content": message.content})
    return req


class LLMApi(ABC):

    @abstractmethod
    def ask(self,
            prompt: Union[str, List[ConversationMessage]],
            param_config: LLMParamConfig = None) -> Union[LLMResp, Iterator]:
        """
        问答
        :param prompt: 单轮提示词或多轮message list
        :param param_config: (Optional)覆盖llm的原本的参数配置，用来控制llm的返回信息
        :return: LLMResp or Iterator(流式打印)
        """

    @abstractmethod
    def set_cache(self, cache: CacheApi):
        """
        设置缓存
        :param cache: 缓存实现对象
        :return: void
        """

    @abstractmethod
    def set_callback(self, callback: BaseCallbackHandler):
        """
        设置Callback回调对象
        :param callback: callback对象
        """

    @abstractmethod
    def ask_for_object(self, prompt: str, class_type: Type[BaseModel], param_config: LLMParamConfig = None):
        """
        问答
        :param prompt: 提示词
        :param class_type: 需要LLM转换的类型
        :param param_config: (Optional)覆盖llm的原本的参数配置，用来控制llm的返回信息
        :return: LLM answer
        """

    @abstractmethod
    def get_llm_config(self) -> LLMConfig:
        """
        获取当前LLM的配置
        :return: LLMConfig
        """


class AbstractLLMApi(LLMApi):

    @staticmethod
    def _build_llm_string(chat_llm: BaseChatModel) -> str:
        return str(sorted([(k, v) for k, v in chat_llm.dict().items()]))

    @staticmethod
    def _parse_llm_response(llm_result: LLMResult) -> LLMResp:
        answer_generation = llm_result.generations[0][0]
        answer = answer_generation.text
        llm_resp = LLMResp(answer=answer, is_from_cache=False)
        return llm_resp

    def __init__(self, llm_config: LLMConfig, chat_llm: BaseChatModel = None, cache: CacheApi = None):
        """ 构造器

          Args:
              llm_config: 通过构造器传递的LLM参数
              chat_llm: （Optional）内部封装的Langchain BaseChatModel的实现类
              cache: CacheApi的实现类，为LLM增加缓存
        """
        self.llm_config = llm_config
        self.chat_llm = chat_llm if chat_llm else self.create_chat_llm_with()

        self.callback_handler = None
        self.cache = cache

    def ask(self,
            prompt: Union[str, List[ConversationMessage]],
            param_config: LLMParamConfig = None) -> Union[LLMResp, Iterator]:
        # 根据入参LLMParamConfig判断是否是流式打印，返回类型为Iterator[str]
        if param_config and param_config.stream:
            return self._stream(prompt=prompt, param_config=param_config)

        # 根据构造器传递的LLMConfig中的LLMParamConfig判断是否是流式打印
        if self.llm_config and self.llm_config.llm_param_config and self.llm_config.llm_param_config.stream:
            return self._stream(prompt=prompt)

        # 非流式打印返回类型为LLMResp
        local_llm = self.chat_llm if param_config is None else self.create_chat_llm_with(param_config)

        # 尝试从缓存获取响应
        if self.cache and isinstance(prompt, str):
            """移除session_tag的构造，session_tag在cache初始化时传入，不在lookup中传递
            """
            cached_llm_rsp = self.cache.lookup(prompt=prompt)
            if cached_llm_rsp:
                """这里需要复制所有的成员变量，当前只有answer和is_from_cache"""
                logger.debug("Hit cached completion. prompt: %s", prompt)
                return LLMResp(answer=cached_llm_rsp.answer, is_from_cache=True)

            logger.debug("Miss cache")

        llm_result: LLMResult = local_llm.generate([self._get_messages(prompt)],
                                                   callbacks=[self.callback_handler] if self.callback_handler else None)

        llm_resp = self.parse_llm_response(llm_result)
        logger.info("resp: %s", llm_resp.answer)
        # 尝试更新缓存
        if self.cache and not llm_resp.is_from_cache and isinstance(prompt, str):
            self.cache.update(prompt=prompt,
                              value=llm_resp)
            logger.debug("Update cache for prompt %s", prompt)

        return llm_resp

    def ask_for_object(self, prompt: str, class_type: Type[BaseModel], param_config: LLMParamConfig = None):
        final_prompt = PromptTemplates.get("system_out_put_parser").format(schema=dict(
            properties=class_type.schema()["properties"]), prompt=prompt)
        llm_resp = self.ask(final_prompt, param_config)

        # 流式打印拼凑结果后输出
        if isinstance(llm_resp, Iterator):
            actual_tokens = []
            for token in llm_resp:
                actual_tokens.append(token)
            answer = "".join(actual_tokens)
        else:
            answer = llm_resp.answer
        return class_type(**json.loads(answer))

    def set_cache(self, cache: CacheApi):
        self.cache = cache

    def default_create_chat_llm_func(self, param_config: LLMParamConfig) -> BaseChatModel:
        """创建chat_llm的默认方法
        使用chat_llm的类直接构造，调用这个默认实现是self.chat_llm必须存在
        使用param_config构造新的chat_llm
        :param param_config: chat_llm参数配置
        :return: 使用新参数构造的chat_llm
        """
        if not self.chat_llm:
            raise ValueError("the default create chat_llm func need preset a BaseChatModel instance")
        llm_type = type(self.chat_llm)
        return llm_type(**param_config.dict())

    def set_callback(self, callback: BaseCallbackHandler):
        self.callback_handler = callback

    def _stream(self,
                prompt: Union[str, List[ConversationMessage]],
                param_config: LLMParamConfig = None) -> Iterator[str]:
        local_llm = self.chat_llm if param_config is None else self.create_chat_llm_with(param_config)
        # 尝试从缓存获取响应
        if self.cache and isinstance(prompt, str):
            """缓存命中，直接从缓存中读取，返回size为1的迭代器
            """
            cached_llm_rsp = self.cache.lookup(prompt=prompt)
            if cached_llm_rsp:
                """返回命中cache"""
                logger.debug("Hit cached completion. prompt: %s", prompt)
                yield cached_llm_rsp.answer
                return
            logger.debug("Miss cache")
        # 通过stream方式获取response

        tokens = local_llm.stream(self._get_messages(prompt),
                                  config=dict(callbacks=[self.callback_handler]) if self.callback_handler else None)

        answer = ""
        for token in tokens:
            yield token.content
            answer += token.content

        logger.info("resp: %s", answer)
        llm_resp = LLMResp(answer=answer, is_from_cache=False)

        # 尝试更新缓存
        if self.cache and not llm_resp.is_from_cache and isinstance(prompt, str):
            self.cache.update(prompt=prompt,
                              value=llm_resp)
            logger.debug("Update cache for prompt %s", prompt)

    def do_create_chat_llm(self, llm_config: LLMConfig):
        return self.default_create_chat_llm_func(llm_config.llm_param_config)

    def create_chat_llm_with(self, param_config: LLMParamConfig = None) -> BaseChatModel:
        if param_config:
            llm_config: LLMConfig = self.llm_config.copy(deep=True)
            llm_config.llm_param_config = param_config
            return self.do_create_chat_llm(llm_config)
        else:
            return self.do_create_chat_llm(self.llm_config)

    def parse_llm_response(self, llm_result: LLMResult) -> LLMResp:
        return self._parse_llm_response(llm_result)

    def _get_messages(self, prompt: Union[str, List[ConversationMessage]]) -> List[ConversationMessage]:
        messages: List[ConversationMessage] = []
        if self.need_add_new_system_message():
            messages.append(ConversationMessage(role=Role.SYSTEM,
                                                content=self.llm_config.llm_module_config.system_prompt))
        if isinstance(prompt, str):
            messages.append(ConversationMessage(role=Role.USER, content=prompt))
        else:
            for msg in prompt:
                if self.need_add_new_system_message() and msg.role.text == "system":
                    continue
                messages.append(msg)
        return messages

    def get_llm_config(self) -> LLMConfig:
        return self.llm_config

    def need_add_new_system_message(self) -> bool:
        return self.llm_config.llm_module_config.enable_append_system_message and \
            self.llm_config.llm_module_config.system_prompt is not None


def get_llm_params(params: dict) -> dict:
    """
    用于过滤llm api参数
    若配置参数，则返回对应参数字典，否则返回空，用默认参数值
    :param params: llm api支持的参数
    :return: 消息体中的dict数据
    """
    llm_param_config = LLMParamConfig(**params)
    return llm_param_config.dict(exclude_none=True)


class LLMApiAdapter(AbstractLLMApi):
    """LLMApi的适配器
    负责把Langchain的LLM实现类适配到LLMApiAdapter

    Attributes:
        chat_llm: 内部封装的Langchain BaseChatModel的实现类
    """

    def __init__(self, chat_llm: BaseChatModel):
        llm_config = LLMConfig()
        super().__init__(llm_config, chat_llm)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict, Any

from pydantic.v1 import BaseModel, Field


class ToolMetadata(BaseModel):
    """
    工具的元数据信息
    Attributes:
        tool_id: 工具ID
        tool_metadata: 工具元数据，如从ES库中检索到的source字段
    """
    tool_id: str
    tool_metadata: Dict[str, Any] = Field(default={})



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional

from pydantic.v1 import Field

from langchain_pangu.pangukitsappdev.api.common_config import IAMConfig, IAMConfigWrapper, HttpConfig
from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings


class SplitConfig(SdkBaseSettings):
    """
    文档解析相关配置
    Attributes:
        css_url: 文档解析地址
        file_path: 文件路径
        file_name: 文件名
        mode: 拆分模式
        iam_config: 文档解析iam_config
        http_config: split使用http_config
    """

    css_url: Optional[str] = Field(env="sdk.doc.split.css.url")
    file_path: Optional[str] = Field(env="sdk.doc.split.css.filepath")
    file_name: Optional[str] = Field(env="sdk.doc.split.css.filename")
    mode: Optional[int] = Field(default=0, env="sdk.doc.split.css.mode")
    iam_config: IAMConfig = Field(default_factory=IAMConfigWrapper(env_prefix="sdk.doc.split.css.iam").get_iam_config)
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.doc.split.css.proxy"))

    def upload_url(self):
        return self.css_url + '/doc-search/files'

    def result_url(self, task_id):
        return f'{self.css_url}/doc-search/tasks/{task_id}'



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional, Type, Union, Callable, Any, get_origin, get_args

from pydantic.v1 import BaseModel, create_model, Field
from pydantic.v1.tools import schema_of
from pydantic.v1.typing import NoneType

from langchain_pangu.pangukitsappdev.api.tool.base import AbstractTool, PanguFunction, DEFAULT_SINGLE_ARG
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates


class Tool(AbstractTool):
    tool_input_schema: Optional[str]
    tool_output_schema: Optional[str]
    pangu_function: Optional[str]

    def _run(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Use the tool."""
        if self.func:
            return self.func(*args, **kwargs)
        raise NotImplementedError("Tool does not support sync")

    def get_input_schema(self) -> str:
        if not self.tool_input_schema:
            self.tool_input_schema = self.build_tool_schema(self.input_desc, self.input_type)
        return self.tool_input_schema

    def get_output_schema(self) -> str:
        if not self.tool_output_schema:
            self.tool_output_schema = self.build_tool_schema(self.output_desc, self.output_type)
        return self.tool_output_schema

    def get_pangu_function(self) -> str:
        if not self.pangu_function:
            self.pangu_function = self.build_pangu_function()
        return self.pangu_function

    @staticmethod
    def build_tool_schema(desc: str, class_type: Union[Type, BaseModel]):
        desc = "空" if class_type == NoneType else desc
        if class_type in [int, float, str, bool, list, dict, NoneType]:
            return PromptTemplates.get("agent_tool_simple_schema").format(desc=desc,
                                                                          type=class_type.__name__)
        # 复杂类型返回JSON schema
        return PromptTemplates.get("agent_tool_json_schema"). \
            format(desc=desc, schema={"properties": class_type.schema().get('properties')})

    def build_pangu_function(self):
        # 诺亚新版本模型并未对principle / results进行训练，临时拼接到description后面
        # principle / results不填值
        arguments = self.get_pang_tool_schema(self.input_desc, self.input_type)
        results = self.get_pang_tool_schema(self.output_desc, self.output_type)
        description = (self.get_tool_desc() + "，" + self.principle) if self.principle else self.get_tool_desc()
        return PanguFunction(name=self.get_tool_id(),
                             description=description,
                             arguments=arguments,
                             results=results).json(ensure_ascii=False, exclude_none=True)

    def get_pang_tool_schema(self, desc: str, class_type: Union[Type, BaseModel]) -> Any:
        # 基本类型
        if class_type in [int, float, str, bool]:
            result_schema = create_model(DEFAULT_SINGLE_ARG, arg=(class_type, Field(description=desc))).schema()
        elif get_origin(class_type) in [list, dict]:
            result_schema = schema_of(class_type)
            result_schema.pop("items", None)
            if len(get_args(class_type)) < 1:
                raise ValueError("List or Dict must specify item type")
            item_type = get_args(class_type)[-1]
            if item_type in [int, float, str, bool, NoneType]:
                result_schema.update({"items": {"type": item_type.__name__}})
            else:
                result_schema["items"] = self.get_pang_tool_schema(desc=desc, class_type=item_type)
        elif class_type is NoneType:
            result_schema = schema_of(class_type)
        else:
            # 复杂类型(BaseModel子类)
            result_schema = class_type.schema()

        definitions = result_schema.pop("definitions", {})
        # 移除冗余title
        result_schema.pop('title', None)
        for prop_key in result_schema.get('properties', {}):
            prop = result_schema.get('properties', {}).get(prop_key, {})
            prop.pop('title', None)
            all_of = prop.pop("allOf", [])
            if all_of:
                enum_name = all_of[0].get('$ref', "").split('/')[-1]
                if definitions.get(enum_name, {}).get("enum") and definitions.get(enum_name, {}).get("type"):
                    prop["enum"] = definitions.get(enum_name, {}).get("enum")
                    prop["type"] = definitions.get(enum_name, {}).get("type")
        return result_schema

    @classmethod
    def from_function(
            cls,
            func: Optional[Callable],
            name: str,
            description: str,
            principle: str,
            input_desc: str,
            output_desc: str,
            args_schema: Optional[Type[BaseModel]] = None,
            return_type: Optional[Type] = None
    ) -> AbstractTool:
        """Initialize tool from a function."""
        if func is None:
            raise ValueError("Function must be provided")
        return cls(
            name=name,
            func=func,
            description=description,
            principle=principle,
            input_desc=input_desc,
            output_desc=output_desc,
            args_schema=args_schema,
            return_type=return_type
        )
